# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
import abc
from decimal import ROUND_DOWN, Decimal
from numbers import Number
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
from compiler_config.config import InlineResultsProcessing

from qat.core.config.configure import get_config
from qat.purr.backends.utilities import (
    UPCONVERT_SIGN,
    PositionData,
    SimpleAcquire,
    evaluate_shape,
    software_post_process_discriminate,
    software_post_process_down_convert,
    software_post_process_linear_map_complex_to_real,
    software_post_process_mean,
)
from qat.purr.compiler.devices import (
    ChannelType,
    MaxPulseLength,
    PulseChannel,
    PulseShapeType,
)
from qat.purr.compiler.emitter import InstructionEmitter, QatFile
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    Assign,
    CustomPulse,
    Delay,
    DeviceUpdate,
    FrequencyShift,
    IndexAccessor,
    Instruction,
    PhaseReset,
    PhaseShift,
    PostProcessing,
    PostProcessType,
    Pulse,
    QuantumInstruction,
    Reset,
    ResultsProcessing,
    Sweep,
    Synchronize,
    Variable,
    Waveform,
    calculate_duration,
)
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt
from qat.purr.utils.logger import get_default_logger
from qat.purr.utils.logging_utils import log_duration

log = get_default_logger()


class InstructionExecutionEngine(abc.ABC):
    def __init__(self, model: QuantumHardwareModel = None, startup_engine: bool = True):
        self.model: Optional[QuantumHardwareModel] = None
        if model is not None:
            self.load_model(model, startup_engine)

    def load_model(self, model: QuantumHardwareModel, startup_engine: bool = True):
        """Shuts down the current hardware, loads the new model then restarts."""
        if self.model is not None:
            self.shutdown()
        self.model = model
        if startup_engine:
            self.startup()

    def startup(self):
        """Starts up the underlying hardware or does nothing if already started."""
        return True

    def run_calibrations(self, qubits_to_calibrate=None):
        pass

    @abc.abstractmethod
    def execute(self, instructions: List[Instruction]) -> Dict[str, Any]:
        pass

    @abc.abstractmethod
    def optimize(self, instructions: List[Instruction]) -> List[Instruction]:
        pass

    @abc.abstractmethod
    def validate(self, instructions: List[Instruction]):
        pass

    def shutdown(self):
        """
        Shuts down the underlying hardware when this instance is no longer in use.
        """
        return True


class QuantumExecutionEngine(InstructionExecutionEngine):
    def __init__(
        self,
        model: QuantumHardwareModel = None,
        startup_engine=True,
        max_instruction_len: int = 200000,
    ):
        super().__init__(model, startup_engine)
        self.max_instruction_len = max_instruction_len

    def _model_exists(self):
        if self.model is None:
            raise ValueError("Requires a loaded hardware model.")

    def _process_assigns(self, results, qfile: "QatFile"):
        """
        As assigns are classical instructions they are not processed as a part of the
        quantum execution (right now).
        Read through the results dictionary and perform the assigns directly, return the
        results.
        """

        def recurse_arrays(results_map, value):
            """Recurse through assignment lists and fetch values in sequence."""
            if isinstance(value, List):
                return [recurse_arrays(results_map, val) for val in value]
            elif isinstance(value, Variable):
                if value.name not in results_map:
                    raise ValueError(
                        f"Attempt to assign variable that doesn't exist {value.name}."
                    )

                if isinstance(value, IndexAccessor):
                    return results_map[value.name][value.index]
                else:
                    return results_map[value.name]
            else:
                return value

        assigns = dict(results)
        for assign in qfile.meta_instructions:
            if not isinstance(assign, Assign):
                continue

            assigns[assign.name] = recurse_arrays(assigns, assign.value)

        return {key: assigns[key] for key in qfile.return_.variables}

    def _process_results(self, results, qfile: "QatFile"):
        """
        Process any software-driven results transformation, such as taking a raw
        waveform result and turning it into a bit, or something else.
        """
        results_processing = {
            val.variable: val
            for val in qfile.meta_instructions
            if isinstance(val, ResultsProcessing)
        }

        missing_results = {
            val.output_variable
            for val in qfile.instructions
            if isinstance(val, Acquire) and val.output_variable not in results_processing
        }

        # For any acquires that are raw, assume they're experiment results.
        for missing_var in missing_results:
            results_processing[missing_var] = ResultsProcessing(
                missing_var, InlineResultsProcessing.Experiment
            )

        for inst in results_processing.values():
            target_values = results.get(inst.variable, None)
            if target_values is None:
                raise ValueError(f"Variable {inst.variable} not found in results output.")

            if (
                InlineResultsProcessing.Raw in inst.results_processing
                and InlineResultsProcessing.Binary in inst.results_processing
            ):
                raise ValueError(
                    f"Raw and Binary processing attempted to be applied "
                    f"to {inst.variable}. Only one should be selected."
                )

            # Strip numpy arrays if we're set to do so.
            if InlineResultsProcessing.NumpyArrays not in inst.results_processing:
                target_values = _numpy_array_to_list(target_values)

            # Transform to various formats if required.
            if InlineResultsProcessing.Binary in inst.results_processing:
                target_values = _binary_average(target_values)

            results[inst.variable] = target_values

        return results

    def optimize(self, instructions: List[Instruction]) -> List[Instruction]:
        """Runs optimization passes specific to this hardware."""
        self._model_exists()
        accum_phaseshifts: Dict[PulseChannel, PhaseShift] = {}
        optimized_instructions: List[Instruction] = []
        for instruction in instructions:
            if isinstance(instruction, PhaseShift) and isinstance(
                instruction.phase, Number
            ):
                if accum_phaseshift := accum_phaseshifts.get(instruction.channel, None):
                    accum_phaseshift.phase += instruction.phase
                else:
                    accum_phaseshifts[instruction.channel] = PhaseShift(
                        instruction.channel, instruction.phase
                    )
            elif isinstance(instruction, (Pulse, CustomPulse)):
                quantum_targets = getattr(instruction, "quantum_targets", [])
                if not isinstance(quantum_targets, List):
                    quantum_targets = [quantum_targets]
                for quantum_target in quantum_targets:
                    if quantum_target in accum_phaseshifts:
                        optimized_instructions.append(accum_phaseshifts.pop(quantum_target))
                optimized_instructions.append(instruction)
            elif isinstance(instruction, PhaseReset):
                for channel in instruction.quantum_targets:
                    accum_phaseshifts.pop(channel, None)
                optimized_instructions.append(instruction)
            else:
                optimized_instructions.append(instruction)
        return optimized_instructions

    def validate(self, instructions: List[Instruction]):
        """Validates this graph for execution on the current hardware."""
        validation_settings = get_config().INSTRUCTION_VALIDATION
        self._model_exists()

        if validation_settings.MAX_INSTRUCTION_LENGTH:
            instruction_length = len(instructions)
            if instruction_length > self.max_instruction_len:
                raise ValueError(
                    f"Program too large to be run in a single block on current hardware. "
                    f"{instruction_length} instructions."
                )

        for inst in instructions:
            if (
                validation_settings.ACQUIRE_CHANNEL
                and isinstance(inst, Acquire)
                and not inst.channel.acquire_allowed
            ):
                raise ValueError(
                    f"Cannot perform an acquire on the physical channel with id "
                    f"{inst.channel.physical_channel}"
                )
            if validation_settings.PULSE_DURATION_LIMITS and isinstance(
                inst, (Pulse, CustomPulse)
            ):
                duration = inst.duration
                if isinstance(duration, Number) and duration > MaxPulseLength:
                    raise ValueError(
                        f"Max Waveform width is {MaxPulseLength} s given: {inst.duration} s"
                    )
                elif isinstance(duration, Variable):
                    values = next(
                        iter(
                            [
                                sw.variables[duration.name]
                                for sw in instructions
                                if isinstance(sw, Sweep)
                                and duration.name in sw.variables.keys()
                            ]
                        )
                    )
                    if np.max(values) > MaxPulseLength:
                        raise ValueError(
                            f"Max Waveform width is {MaxPulseLength} s given: {values} s"
                        )

    def _generate_repeat_batches(self, repeats):
        """
        Batches together executions if we have more than the amount the hardware can
        handle at once. A repeat limit of -1 disables this functionality and just runs
        everything at once.
        """
        batch_limit = self.model.repeat_limit
        if batch_limit == -1:
            return [repeats]

        list_expansion_ratio = int(
            (Decimal(repeats) / Decimal(batch_limit)).to_integral(ROUND_DOWN)
        )
        remainder = repeats % batch_limit
        batches = [batch_limit] * list_expansion_ratio
        if remainder > 0:
            batches.append(remainder)

        return batches

    def execute(self, instructions):
        """Executes this qat file against this current hardware."""
        return self._common_execute(instructions)

    def _execute_with_interrupt(self, instructions, interrupt: Interrupt = NullInterrupt()):
        """Executes this qat file against this current hardware.
        Execution allows for interrupts triggered by events
        """
        return self._common_execute(instructions, interrupt)

    def _common_execute(self, instructions, interrupt: Interrupt = NullInterrupt()):
        """Executes this qat file against this current hardware."""
        self._model_exists()

        with log_duration("QPU returned results in {} seconds."):
            # This is where we'd send the instructions off to the compiler for
            # processing, for now do ad-hoc processing.
            qat_file = InstructionEmitter().emit(instructions, self.model)

            # Rebuild repeat list if the hardware can't support the current setup.
            repeat_count = qat_file.repeat.repeat_count

            MAX_REPEATS_LIMIT = get_config().MAX_REPEATS_LIMIT
            if repeat_count > MAX_REPEATS_LIMIT:
                raise ValueError(
                    f"Number of shots {repeat_count} exceeds the maximum amount of {MAX_REPEATS_LIMIT}."
                )

            if repeat_count > self.model.repeat_limit:
                log.info(
                    f"Running {repeat_count} shots at once is "
                    f"unsupported on the current hardware. Batching execution."
                )
            batches = self._generate_repeat_batches(repeat_count)

            # Set up the sweep and device injectors, passing all appropriate data to
            # the specific execute methods and deal with clean-up afterwards.
            results = {}
            for i, batch_count in enumerate(batches):
                metadata = {"batch_iteration": i}
                interrupt.if_triggered(metadata, throw=True)
                if batch_count <= 0:
                    continue

                # Reset iterators/injectors per cycle.
                switerator = SweepIterator.from_qfile(qat_file)
                dinjectors = DeviceInjectors(qat_file.instructions)

                # Assign shortened repeat for this execution then reset it.
                qat_file.repeat.repeat_count = batch_count
                try:
                    dinjectors.inject()
                    batch_results = self._execute_on_hardware(
                        switerator, qat_file, interrupt
                    )
                finally:
                    switerator.revert(qat_file.instructions)
                    dinjectors.revert()
                    qat_file.repeat.repeat_count = repeat_count

                results = self._accumulate_results(results, batch_results)

            # Process metadata assign/return values to make sure the data is in the
            # right form.
            results = self._process_results(results, qat_file)
            results = self._process_assigns(results, qat_file)

            return results

    @staticmethod
    def _accumulate_results(results: Dict, batch_results: Dict):
        if not any(batch_results):
            return results

        if any(results):
            if results.keys() != batch_results.keys():
                raise ValueError(
                    f"Dictionaries' keys mismatch, {results.keys()} != {batch_results.keys()}"
                )

            for key in results:
                existing = results[key]
                appending = batch_results[key]

                if type(existing) is not type(appending):
                    raise ValueError(
                        f"Expected objects with the same type, got {type(existing)} and {type(appending)} instead"
                    )

                if isinstance(existing, np.ndarray):
                    # TODO: Certain postprocessing instructions result in
                    # strange concatenation, e.g. sequential mean.
                    results[key] = np.concatenate(
                        (existing, appending), axis=existing.ndim - 1
                    )
                elif isinstance(existing, List):

                    def combine_lists(exi, new):
                        if len(new) > 0 and not isinstance(new[0], list):
                            exi += new
                            return
                        for i, item in enumerate(new):
                            combine_lists(exi[i], item)

                    combine_lists(existing, appending)
                else:
                    raise ValueError(
                        f"Cannot combine objects of unsupported type {type(existing)}"
                    )
        else:
            results = batch_results

        return results

    @abc.abstractmethod
    def _execute_on_hardware(
        self, sweep_iterator, package: "QatFile", interrupt: Interrupt = NullInterrupt()
    ) -> Union[Dict[str, List[float]], Dict[str, np.ndarray]]:
        pass

    def create_duration_timeline(self, instructions: List[QuantumInstruction]):
        """
        Builds a dictionary mapping channels to a list of instructions and at what
        precise sample time they should be associated with. It's important that the
        times are absolute here, not relative.

        :Example:

        .. code-block:: python

            {"drive": [
                (0, 1000, pulse),
                (1000, 2000, pulse),
                (1000, 2000, acquire)
                ...
            ]}
        """
        results: Dict[PulseChannel, List[PositionData]] = {}
        total_durations: Dict[PulseChannel, int] = dict()

        for instruction in instructions:
            # TODO: Acquire is a special quantum target for post processing.
            #  This should probably be changed.
            qtargets = [
                qtarget.channel if isinstance(qtarget, Acquire) else qtarget
                for qtarget in instruction.quantum_targets
            ]

            # For syncs we want to look at the currently-processed instructions on
            # the channels we target, get the max end time then align all of our
            # channels to that point in time.
            if isinstance(instruction, Synchronize):
                longest_length = max(
                    [total_durations.setdefault(qt, 0) for qt in qtargets],
                    default=0.0,
                )
                for qtarget in qtargets:
                    device_instructions: List[PositionData] = results.setdefault(
                        qtarget, []
                    )
                    delay_time = longest_length - total_durations[qtarget]
                    if delay_time > 0:
                        start = (
                            device_instructions[-1].end if any(device_instructions) else 0
                        )
                        instr = Delay(qtarget, delay_time)
                        position_data = PositionData(
                            start,
                            start + calculate_duration(instr),
                            instr,
                        )
                        device_instructions.append(position_data)
                        total_durations[qtarget] = longest_length
            else:
                duration = calculate_duration(instruction)
                for qtarget in qtargets:
                    device_instructions: List[PositionData] = results.setdefault(
                        qtarget, []
                    )
                    start = device_instructions[-1].end if any(device_instructions) else 0
                    device_instructions.append(
                        PositionData(
                            start,
                            start + duration,
                            instruction,
                        )
                    )
                    total_durations[qtarget] = (
                        total_durations.get(qtarget, 0) + instruction.duration
                    )

        # Strip timelines that only hold delays and phase resets, since that just means nothing is
        # happening on this channel.
        results = {
            key: timeline
            for key, timeline in dict(results).items()
            if not all(
                isinstance(pos_data.instruction, (Delay, PhaseReset, PhaseShift))
                for pos_data in timeline
            )
        }

        if final_positions := [
            final_position[-1].end for final_position in list(results.values())
        ]:
            end_point = max(final_positions)

            for qubit in self.model.qubits:
                freq_channel = next(
                    (
                        channel
                        for channel in qubit.get_all_channels()
                        if channel.channel_type == ChannelType.freq_shift
                    ),
                    None,
                )

                if freq_channel is not None:
                    if freq_channel.active:
                        pulse = Pulse(
                            freq_channel,
                            shape=PulseShapeType.SQUARE,
                            amp=freq_channel.amp,
                            phase=getattr(freq_channel, "phase", 0),
                            width=end_point * qubit.physical_channel.sample_time,
                        )
                        results[freq_channel] = [
                            PositionData(0, end_point, pulse),
                        ]
        return results

    def build_pulse_channel_buffers(
        self,
        position_map: Dict[PulseChannel, List[PositionData]],
        do_upconvert: bool = True,
    ):
        buffers = {}
        for pulse_channel, positions in position_map.items():
            buffers[pulse_channel] = buffer = np.zeros(
                positions[-1].end if any(positions) else 0, dtype=np.complex128
            )

            phase = 0.0
            frequency = pulse_channel.frequency
            for pos in positions:
                if isinstance(pos.instruction, Waveform):
                    phase = self.process_pulse(
                        pos, buffer, pulse_channel, phase, frequency, do_upconvert
                    )
                elif isinstance(pos.instruction, PhaseShift):
                    phase = self.process_phaseshift(pos, phase)
                elif isinstance(pos.instruction, PhaseReset):
                    phase = self.process_phasereset(pos, phase)
                elif isinstance(pos.instruction, Reset):
                    self.process_reset(pos)
                elif isinstance(pos.instruction, FrequencyShift):
                    frequency = self.process_frequencyshift(pos, frequency, pulse_channel)

        return buffers

    def process_pulse(
        self,
        position: PositionData,
        buffer: List[float],
        pulse_channel: PulseChannel,
        phase: float,
        frequency: float,
        do_upconvert: bool,
    ):
        dt = pulse_channel.sample_time

        samples = position.end - position.start
        length = (position.end - position.start) * dt
        centre = length / 2.0
        t = np.linspace(
            start=-centre + 0.5 * dt, stop=length - centre - 0.5 * dt, num=samples
        )
        pulse = evaluate_shape(position.instruction, t, phase)

        scale = pulse_channel.scale
        if (
            isinstance(position.instruction, (Pulse, CustomPulse))
            and position.instruction.ignore_channel_scale
        ):
            scale = 1
        pulse *= scale
        pulse += pulse_channel.bias

        if do_upconvert:
            t += centre - 0.5 * dt + position.start * dt
            pulse = self.do_upconvert(pulse, t, pulse_channel, frequency)

        buffer[position.start : position.end] = pulse

        return phase

    def process_phaseshift(self, position: PositionData, phase: float):
        return phase + position.instruction.phase

    def process_phasereset(self, position: PositionData, phase: float):
        """Child hardware types might need to know the phase and instruction"""
        return 0

    def process_reset(self, position: PositionData):
        pass

    def process_frequencyshift(
        self, position: PositionData, frequency: float, pulse_channel: PulseChannel
    ):
        # Check no pulse channels on this physical channel used a fixed if
        for channel in self.model.get_pulse_channels_from_physical_channel(
            pulse_channel.physical_channel
        ):
            if channel.fixed_if:
                raise NotImplementedError(
                    "Hardware does not currently support frequency shifts on the "
                    f"physical channel {pulse_channel.physical_channel}."
                )

        new_frequency = frequency + position.instruction.frequency
        if (
            new_frequency < pulse_channel.min_frequency
            or new_frequency > pulse_channel.max_frequency
        ):
            raise ValueError(
                f"Cannot shift pulse channel frequency from '{frequency}' to "
                f"'{new_frequency}'. The new frequency must be between the bounds "
                f"({pulse_channel.min_frequency}, {pulse_channel.max_frequency}) on "
                f"physical channel with id {pulse_channel.physical_channel_id}."
            )

        return new_frequency

    def do_upconvert(
        self,
        buffer: List[float],
        time: List[float],
        pulse_channel: PulseChannel,
        frequency: float,
    ):
        tslip = pulse_channel.phase_offset
        imbalance = pulse_channel.imbalance
        if pulse_channel.fixed_if:
            freq = pulse_channel.baseband_if_frequency
        else:
            freq = frequency - pulse_channel.baseband_frequency
        buffer *= np.exp(UPCONVERT_SIGN * 2.0j * np.pi * freq * time)
        if not tslip == 0.0:
            buffer_slip = buffer * np.exp(UPCONVERT_SIGN * 2.0j * np.pi * freq * tslip)
            buffer.imag = buffer_slip.imag
        if not imbalance == 1.0:
            buffer.real /= imbalance**0.5
            buffer.imag *= imbalance**0.5

        return buffer

    def build_physical_channel_buffers(
        self, pulse_channel_buffers: Dict[PulseChannel, np.ndarray]
    ):
        # Add all pulse channel buffers belonging to the same physical channel together
        buffers = {}
        for physical_channel_id in self.model.physical_channels.keys():
            physical_channel_pulse_buffers = [
                buffer
                for pulse_channel, buffer in pulse_channel_buffers.items()
                if pulse_channel.physical_channel_id == physical_channel_id
            ]
            physical_channel_buffer = buffers.setdefault(
                physical_channel_id,
                np.zeros(
                    max(map(len, physical_channel_pulse_buffers), default=0),
                    dtype=np.complex128,
                ),
            )
            for buffer in physical_channel_pulse_buffers:
                physical_channel_buffer[0 : len(buffer)] += buffer

        return buffers

    def build_acquire_list(self, position_map: Dict[PulseChannel, List[PositionData]]):
        buffers = {}
        for pulse_channel, positions in position_map.items():
            buffer = buffers.setdefault(pulse_channel.full_id(), [])
            for pos in positions:
                if isinstance(pos.instruction, Acquire):
                    if pos.instruction.filter is not None:
                        raise NotImplementedError(
                            "Hardware currently doesn't support custom filters."
                        )

                    samples = pos.end - pos.start
                    buffer.append(
                        SimpleAcquire(
                            pos.start,
                            samples,
                            pos.instruction.output_variable,
                            pulse_channel,
                            pulse_channel.physical_channel,
                            pos.instruction.mode,
                            pos.instruction.delay,
                            pos.instruction,
                        )
                    )

        return buffers

    def run_post_processing(self, post_processing: PostProcessing, value, value_axis):
        """Run post-processing on the response via software."""
        if post_processing.process == PostProcessType.DOWN_CONVERT:
            return software_post_process_down_convert(
                post_processing.args, post_processing.axes, value, value_axis
            )
        elif post_processing.process == PostProcessType.MEAN:
            return software_post_process_mean(post_processing.axes, value, value_axis)
        elif post_processing.process == PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL:
            return software_post_process_linear_map_complex_to_real(
                post_processing.args, value, value_axis
            )
        elif post_processing.process == PostProcessType.DISCRIMINATE:
            return software_post_process_discriminate(
                post_processing.args, value, value_axis
            )

    @property
    def error_mitigation(self):
        return self.model.error_mitigation

    def __repr__(self):
        if self.model is not None:
            return f"{self.__class__.__name__} with {len(self.model.qubits)} qubits"
        return "No hardware loaded."


def _complex_to_binary(number: complex):
    """Base calculation for changing a complex measurement to binary form."""
    return 0 if number.real > 0 else 1


def _binary(results_list):
    """Changes all measurements to binary format."""
    if not isinstance(results_list, Iterable):
        return [results_list]

    results = []
    for item in results_list:
        if isinstance(item, Iterable):
            results.append(_binary(item))
        elif isinstance(item, complex):  # If we're a flat register, just append.
            results.append(_complex_to_binary(item))
        elif isinstance(item, float):
            results.append(0 if item > 0 else 1)
        else:
            results.append(item)
    return results


def _binary_average(results_list):
    """
    Averages all repeat results and returns a definitive 1/0 for each qubit measurement.
    """
    # If we have many sweeps/repeats loop through all of them and sum.
    if all([isinstance(val, list) for val in results_list]):
        binary_results = [_binary_average(nested) for nested in results_list]
    else:
        binary_results = _binary(results_list)

    return 1 if sum(binary_results) >= (len(binary_results) / 2) else 0


def _numpy_array_to_list(array):
    """Transform numpy arrays to a normal list."""
    if isinstance(array, np.ndarray):
        numpy_list: List = array.tolist()
        if len(numpy_list) == 1:
            return numpy_list[0]
        return numpy_list
    elif isinstance(array, List):
        list_list = [_numpy_array_to_list(val) for val in array]
        if len(list_list) == 1:
            return list_list[0]
        return list_list
    else:
        return array


class DeviceInjector:
    """
    Injects a value into a device for the entirety of a sweeps duration. Analogous to
    setting a device value before execution.
    """

    def __init__(self, target: DeviceUpdate):
        super().__init__()
        self.updater = target
        self.revert_value = None

    def inject(self):
        if isinstance(self.updater.target, Dict):
            self.revert_value = self.updater.target[self.updater.attribute]
            self.updater.target[self.updater.attribute] = self.updater.value
        else:
            self.revert_value = getattr(self.updater.target, self.updater.attribute)
            setattr(self.updater.target, self.updater.attribute, self.updater.value)

    def revert(self):
        if isinstance(self.updater.target, Dict):
            self.updater.target[self.updater.attribute] = self.revert_value
        else:
            setattr(self.updater.target, self.updater.attribute, self.revert_value)

    def __repr__(self):
        return (
            f"{self.updater.target.id}.{self.updater.attribute} = {str(self.updater.value)}"
        )


class ValueReplacement(abc.ABC):
    """Abstract class for values that inject during a sweep."""

    @abc.abstractmethod
    def replace(self, field_value, symbols, index):
        pass

    @abc.abstractmethod
    def revert(self):
        pass


class VariableInjector(ValueReplacement):
    """
    All injected values are assumed to be in a list of len(sweep_length), this just
    fetches out the particular value of a list to inject as we're sweeping.

    So calling this with ``name='frequency', index=5``, from:

    .. code-block:: python

        {
            'frequency': [1, 2, 4, 8, 16, 32, 64, ...],
            ...
        }

    Would inject 32 into this field, as we're on the 6th iteration of this sweep.
    """

    def __init__(self, var):
        self.var: Variable = var

    def replace(self, field_value, replacements, index):
        if self.var.name in replacements:
            return replacements[self.var.name][index]
        else:
            return field_value

    def revert(self):
        return self.var


class IteratorInjector(ValueReplacement):
    """
    Value injector for objects in lists, tuples and dictionaries.

    Currently when we inject an iteration into an iteration of the same type - resolving
    a ValueReplacer entry in a dictionary also returns a dictionary - it merges that
    value into the iteration instead of a simple insertion.

    So the list ``[1, 2, ValueReplacer(...)]`` that resolves into ``[1, 2, [3, 4, 5]]`` will become ``[1, 2, 3, 4, 5]``. But
    if it was ``[1, 2, (3, 4, 5)]`` it will stay as such.
    """

    def __init__(self, iteration):
        self.iteration = iteration

    def replace(self, field_value, replacements, index):
        if isinstance(self.iteration, (List, Tuple)):
            type_to_check = type(self.iteration)
            expanded = []
            for val in self.iteration:
                if isinstance(val, Variable):
                    replacement_value = replacements[val.name][index]
                    if isinstance(replacement_value, type_to_check):
                        expanded.extend(replacement_value)
                    else:
                        expanded.append(replacement_value)
                else:
                    expanded.append(val)
            return type_to_check(expanded)
        elif isinstance(self.iteration, Dict):
            return {
                key: (
                    replacements[value.name][index]
                    if isinstance(value, Variable) and key in replacements
                    else value
                )
                for key, value in self.iteration.items()
            }
        else:
            return self.iteration

    def revert(self):
        return self.iteration


class InjectionMetadata:
    """
    A class which is injected into an instruction to hold data about its state before
    injection started, as well as objects that facilitate dynamic injection of values
    into fields.

    It'll replace fields in the object that hold ``Variable('X')`` with the value of ``X``. When revert is called it replaces
    the field with ``Variable('X')`` again, essentially reseting it for next execution.
    """

    field: str = "_$injection_metadata"

    def __init__(self):
        self.variables: Dict[str, ValueReplacement] = dict()

    def inject(self, node, replacements: Dict, index):
        for field, replacer in self.variables.items():
            setattr(
                node, field, replacer.replace(getattr(node, field), replacements, index)
            )

    def revert(self, node):
        for field, replacer in self.variables.items():
            setattr(node, field, replacer.revert())

    def is_empty(self):
        return len(self.variables) == 0


class DeviceInjectors:
    """
    Special sort of injector that is a sort of double-injection. It takes an object that
    needs to be modified during a sweep, replaces certain fields with Variable (and
    others) and lets normal injection do it's work, then at the end of the sweep resets
    the field with the original value.
    """

    def __init__(self, instructions):
        self.injectors = [
            DeviceInjector(val) for val in instructions if isinstance(val, DeviceUpdate)
        ]

    def inject(self):
        for dinject in self.injectors:
            dinject.inject()

    def revert(self):
        for dinject in self.injectors:
            dinject.revert()


class SweepIterator:
    """
    Acts as a controller for sweep-reliant values and iterates the amount of times
    designated by the sweep.

    Every time ``do_sweep`` is called it will inject the current sweep values into the variable, so if passed
    a sweep list with ``{'dave': [1, 2, 3, 4, 5]}`` it will iterate 5 times and inject the values ``1...5`` into
    objects that has ``Variable('dave')``.

    Nested sweeps act as inner loops, and iterates inner to outer. So with a nested sweep of ``{'sam': [5, 10, 15]}``
    and using our previous example we will sweep a total of 15 times, looking a little like this:

    .. code-block:: python

        dave = 1, sam = 5
        dave = 1, sam = 10
        dave = 1, sam = 15
        dave = 2, sam = 5
        ...
    """

    def __init__(self, sweep=None, nested_sweep=None):
        self.sweep: Sweep = sweep
        self.current_iteration = -1
        self.nested_sweep: "SweepIterator" = nested_sweep

        # An injector is an object that does more complicated value injection with a
        # revert at the end.
        self.injection_match_prefixs = {
            SweepIterator.__module__.split(".")[0],
            self.__module__.split(".")[0],
        }

    @property
    def accumulated_sweep_iteration(self):
        """
        The current number of total iterations this sweep as run as a whole, including
        nested sweeps.
        """
        if self.nested_sweep is not None:
            return (
                self.current_iteration * self.nested_sweep.length
                + self.nested_sweep.accumulated_sweep_iteration
            )
        else:
            return self.current_iteration + 1

    def get_current_sweep_iteration(self):
        """
        Returns the current iteration of the entire loop nest seen as a polyhedra.
        The returned structure is a list representing a lattice point p where p[i]
        indicates the current iteration of the sweep at level i
        """
        if self.sweep is None:
            return []
        elif self.nested_sweep is None:
            return [self.current_iteration]
        else:
            return [
                self.current_iteration
            ] + self.nested_sweep.get_current_sweep_iteration()

    @staticmethod
    def from_qfile(qfile: "QatFile"):
        """Build a sweep iterator from a .qat file."""
        sweep_iterator = SweepIterator()
        for meta in qfile.meta_instructions:
            if isinstance(meta, Sweep):
                sweep_iterator.add_sweep(meta)

        return sweep_iterator

    @staticmethod
    def from_sweeps(sweeps: List[Sweep]):
        """
        Convenient helper that builds a SweepIterator from a collection of sweeps.
        This bypasses QatFile and is more suitable for the TriagePass.
        """

        sweep_iterator = SweepIterator()
        for sweep in sweeps:
            sweep_iterator.add_sweep(sweep)
        return sweep_iterator

    def _revert_sweep_values(self, node):
        """
        Reverts all changes made by injectors during running to their pre-sweep states.
        The execution should be able to be re-run using the same Sweep object and
        results should be the same.
        """
        # TODO: Only a very small amount of code is needed by the revert, pull out into
        #   new method.
        self._inject_sweep_values(node, None, None, revert=True)

    def _inject_sweep_values(
        self, node, replacers, index, revert=False, recursion_guard=None
    ):
        """
        Injects values into objects, mostly making sure that Variable objects turn into
        their actual values.
        """
        # If we can't get its internal objects, skip.
        has_dict = getattr(node, "__dict__", None)
        if not has_dict:
            return

        if recursion_guard is None:
            recursion_guard = set()

        # If we've already seen this object and it's a recursive call, skip.
        if id(node) in recursion_guard:
            return

        # If we're not an object we can check the location of, skip.
        module = type(node).__dict__.get("__module__", None)
        if module is None or (
            self.injection_match_prefixs is not None
            and not any(module.startswith(val) for val in self.injection_match_prefixs)
        ):
            return

        # Add our object to the recursion guard so we don't recurse.
        recursion_guard.add(id(node))

        # Get injection metadata, and if don't have any generate it.
        injection_meta = getattr(node, InjectionMetadata.field, None)
        if injection_meta is None:
            injection_meta = InjectionMetadata()
            for field, value in filter(
                lambda kvp: not kvp[0].startswith("_"), node.__dict__.items()
            ):
                if isinstance(value, Variable):
                    injection_meta.variables[field] = VariableInjector(value)
                elif (
                    isinstance(value, (Tuple, List))
                    and any([isinstance(val, Variable) for val in value])
                ) or (
                    isinstance(value, Dict)
                    and any([isinstance(val, Variable) for val in value.values()])
                ):
                    injection_meta.variables[field] = IteratorInjector(value)

            setattr(node, InjectionMetadata.field, injection_meta)

        # Perform injection on this node, then recurse into all objects taking into
        # account iterables.
        if revert:
            injection_meta.revert(node)
            delattr(node, InjectionMetadata.field)
        else:
            injection_meta.inject(node, replacers, index)

        for value in node.__dict__.values():
            if isinstance(value, InjectionMetadata):
                continue

            if isinstance(value, (Tuple, List, Dict)):
                for val in value:
                    self._inject_sweep_values(
                        val, replacers, index, revert, recursion_guard
                    )
            else:
                self._inject_sweep_values(value, replacers, index, revert, recursion_guard)

    def reset_iteration(self):
        """
        Resets iteration of this sweep and all children since we infer that if we've
        finished, then our children have too.
        """
        self.current_iteration = -1
        if self.nested_sweep is not None:
            self.nested_sweep.reset_iteration()

    def do_sweep(self, instructions):
        """
        Start/continue sweeping. Injects values into the instructions appropriate for
        this sweep.
        """
        if self.nested_sweep is not None:
            # Special-case where a parent sweep won't increment until its child is
            # complete, but we start at -1 so need to bump it forward a single value.
            if self.current_iteration == -1:
                self.current_iteration = 0

            if self.nested_sweep.is_finished():
                self.current_iteration += 1
                self.nested_sweep.reset_iteration()
        else:
            self.current_iteration += 1

        if self.sweep is None:
            return

        for inst in instructions:
            self._inject_sweep_values(inst, self.sweep.variables, self.current_iteration)

        if self.nested_sweep is not None:
            self.nested_sweep.do_sweep(instructions)

    def add_sweep(self, nested: Sweep):
        """
        Adds a sweep onto this iterator. If it already has a sweep, adds it as a nested
        one.
        """
        if self.sweep is None:
            self.sweep = nested
        elif self.nested_sweep is None:
            self.nested_sweep = SweepIterator(nested)
        else:
            self.nested_sweep.add_sweep(nested)

    def revert(self, instructions):
        """Revert all instruction modifications to their pre-sweep state."""
        if self.sweep is None:
            return

        for inst in instructions:
            self._revert_sweep_values(inst)

        if self.nested_sweep is not None:
            self.nested_sweep.revert(instructions)

    def is_finished(self):
        """Is this sweep finished."""
        # We're zero-index so take that into account.
        return self.current_iteration == (
            self.sweep.length - 1 if self.sweep is not None else 0
        ) and (self.nested_sweep is None or self.nested_sweep.is_finished())

    def get_results_shape(self, shape: Tuple = None):
        """Return a default array that mirrors the structure for this set of sweeps."""
        if self.nested_sweep is not None:
            return (self.sweep.length, *self.nested_sweep.get_results_shape(shape))
        return (self.length, *shape) if shape is not None else (self.length,)

    @property
    def length(self):
        """Returns actual length of the entire sweep, so

        ``sweep_length * nested_length * nested_length ...``
        """
        if self.sweep is None:
            return 1

        return self.sweep.length * (
            self.nested_sweep.length if self.nested_sweep is not None else 1
        )

    def insert_result_at_sweep_position(self, results_array: np.array, value: np.array):
        """Insert the response value into the appropriate nested array."""
        if self.nested_sweep is not None:
            self.nested_sweep.insert_result_at_sweep_position(
                results_array[self.current_iteration], value
            )
        else:
            results_array[self.current_iteration] = value



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import math
from copy import deepcopy
from functools import singledispatchmethod
from importlib.util import find_spec
from numbers import Number
from os.path import join
from pathlib import Path
from pydoc import locate
from typing import Any

import numpy as np
from compiler_config.config import InlineResultsProcessing, Languages
from lark import Lark, Token, Tree, UnexpectedCharacters
from lark.visitors import Interpreter
from openqasm3 import ast
from openqasm3.parser import parse as oq3_parse
from openqasm3.visitor import QASMVisitor

from qat.frontend.parsers.qasm.base import AbstractParser, ParseResults, QasmContext
from qat.frontend.register import BitRegister, CregIndexValue, QubitRegister, Registers
from qat.ir.instruction_builder import QuantumInstructionBuilder
from qat.ir.instructions import (
    FrequencyShift,
    PhaseReset,
    PhaseShift,
    QuantumInstruction,
    Variable,
)
from qat.ir.measure import Acquire, AcquireMode, PostProcessType, ProcessAxis
from qat.ir.pulse_channel import PulseChannel as IRPulseChannel
from qat.ir.waveforms import (
    AbstractWaveform,
    DragGaussianWaveform,
    GaussianSquareWaveform,
    GaussianWaveform,
    GaussianZeroEdgeWaveform,
    Pulse,
    RoundedSquareWaveform,
    SampledWaveform,
    SechWaveform,
    SinWaveform,
    SoftSquareWaveform,
    SquareWaveform,
    Waveform,
)
from qat.model.device import (
    AcquirePulseChannel,
    PhysicalChannel,
    PulseChannel,
    Qubit,
    QubitPulseChannels,
    ResonatorPulseChannels,
)
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.utils.pydantic import ValidatedList

log = get_default_logger()


class OpenPulseContext(QasmContext):
    calibration_methods: dict[str, Any] = dict()


def get_frame_mappings(hw: PhysicalHardwareModel):
    """
    Generate the names for frames we allow in open pulse 'extern' statements.
    Returns a dictionary mapping name->pulse channel.
    """
    frames = {}
    for q1_id, q1 in hw.qubits.items():
        for channel_type in QubitPulseChannels.model_fields:
            if "cross_resonance" in channel_type:
                for q2_id, pulse_channel in getattr(
                    q1.pulse_channels, channel_type
                ).items():
                    qubit_prefix = f"q{q1_id}_q{q2_id}"
                    channel_type = channel_type.replace("_channels", "")
                    frames[f"{qubit_prefix}_{channel_type}"] = pulse_channel
            else:
                qubit_prefix = f"q{q1_id}"
                frames[f"{qubit_prefix}_{channel_type}"] = getattr(
                    q1.pulse_channels, channel_type
                )

        for channel_type in ResonatorPulseChannels.model_fields:
            res_prefix = f"r{q1_id}"
            frames[f"{res_prefix}_{channel_type}"] = getattr(
                q1.resonator.pulse_channels, channel_type
            )

    return frames


def get_port_mappings(hw: PhysicalHardwareModel):
    """
    Generate the names for ports we allow in open pulse 'extern' statements.
    Returns a dictionary mapping name->physical channel.
    """
    ports = {}
    for index, channel in hw.physical_channel_map.items():
        name = f"channel_{index}"
        ports[name] = channel

    return ports


class Qasm3ParserBase(AbstractParser, QASMVisitor):
    def parser_language(self) -> Languages:
        return Languages.Qasm3

    def _includes_standard_gates(self, qasm_str: str) -> bool:
        return any(
            [
                include in qasm_str
                for include in ('include "qelib1.inc";', 'include "stdgates.inc";')
            ]
        )

    def load_default_gates(self, context: QasmContext) -> QasmContext:
        node = ast.Include(filename="stdgates.inc")
        self.visit(node, context)
        return context

    def parse(
        self, builder: QuantumInstructionBuilder, qasm: str
    ) -> QuantumInstructionBuilder:
        self.builder = builder
        # Parse or pick up the cached version, then remove it as we are about to
        # interpret it.
        program = self._fetch_or_parse(qasm)
        if (qasm_id := hash(qasm)) in self._cached_parses:
            del self._cached_parses[qasm_id]
        context = QasmContext()
        if not self._includes_standard_gates(qasm):
            self.load_default_gates(context)
        return self.process_program(program, context)

    def process_program(
        self, prog: ast.Program, context: QasmContext | None
    ) -> QuantumInstructionBuilder:
        context = context or QasmContext()
        self._walk_program(prog, context)
        self._assign_returns(context)
        return self.builder

    def _fetch_or_parse(self, qasm_str: str) -> ast.Program:
        if 'defcalgrammar "openpulse"' in qasm_str:
            raise ValueError("QASM3ParserBase can not parse OpenPulse programs.")

        # If we have seen this file before.
        qasm_id = hash(qasm_str)
        if (cached_value := self._cached_parses.get(qasm_id, None)) is not None:
            return cached_value

        try:
            program = oq3_parse(qasm_str)
        except Exception as e:
            invalid_string = str(e)
            raise ValueError(f"Invalid QASM 3 syntax: '{invalid_string}'.")

        self._cached_parses[qasm_id] = program
        return program

    def validate(self, prog: ast.Program):
        pass

    def modify(self, prog: ast.Program):
        """
        Allows children to transform the program before validation/transforming into our
        AST occurs.
        """
        pass

    def _walk_program(self, prog: ast.Program, context: QasmContext):
        self.modify(prog)
        self.validate(prog)

        for node in prog.statements:
            self.visit(node, context)

    def visit_Include(self, node: ast.Include, context: QasmContext):
        if node.filename == "qelib1.inc":
            file_path = Path(
                find_spec("qiskit.qasm.libs").submodule_search_locations[0], node.filename
            )
        elif node.filename == "stdgates.inc":
            file_path = Path(Path(__file__).parents[1], "grammars", node.filename)
        else:
            file_path = Path(node.filename)
        if not file_path.is_file():
            raise ValueError(f"File not found for '{str(file_path)}'.")
        with file_path.open(encoding="utf-8") as f:
            self._walk_program(oq3_parse(f.read()), context)

    def visit_QuantumGateDefinition(
        self, node: ast.QuantumGateDefinition, context: QasmContext
    ):
        context.gates[node.name.name] = node

    @singledispatchmethod
    def _get_qubits(self, input_, context: QasmContext):
        raise NotImplementedError(f"Cannot get qubits from {type(input_)}")

    @_get_qubits.register(list)
    def _(self, input_, context: QasmContext):
        qubits = []
        for item in input_:
            qbs = self._get_qubits(item, context)
            if isinstance(qbs, list):
                qubits.extend(qbs)
            else:
                qubits.append(qbs)
        return qubits

    @_get_qubits.register(ast.QASMNode)
    def _(self, input_, context: QasmContext):
        return self.visit(input_, context)

    @_get_qubits.register(Qubit)
    @_get_qubits.register(QubitRegister)
    def _(self, input_, context: QasmContext):
        return input_

    def _create_qb_specific_gate_suffix(
        self, name: str, target_qubits: list[Qubit | QubitRegister]
    ) -> str:
        return f"{name}[{','.join([str(qb) for qb in target_qubits])}]"

    def _attempt_declaration(self, variable: Variable, context: QasmContext):
        if (old_var := context.variables.get(variable.name, None)) is None:
            context.variables[variable.name] = variable
        elif old_var != variable:
            raise ValueError(f"Can't redeclare variable {variable.name}")

    def visit_ClassicalDeclaration(
        self, node: ast.ClassicalDeclaration, context: QasmContext
    ):
        self.add_creg(node.identifier.name, node.type.size.value, context)

    def visit_QubitDeclaration(self, node: ast.QubitDeclaration, context: QasmContext):
        self.add_qreg(node.qubit.name, node.size.value, context, self.builder)

    def _get_node_arguments(self, node: ast.QASMNode, context: QasmContext) -> list:
        args = []
        for arg in node.arguments:
            if isinstance(arg, ast.QASMNode):
                args.append(self.visit(arg, context))
            else:
                args.append(arg)
        return args

    def visit_BinaryExpression(self, node: ast.BinaryExpression, context: QasmContext):
        lhs = self.visit(node.lhs, context)
        rhs = self.visit(node.rhs, context)
        op = node.op.name

        match op:
            case "+":
                return lhs + rhs
            case "-":
                return lhs - rhs
            case "/":
                return lhs / rhs
            case "*":
                return lhs * rhs
            case ">":
                return lhs > rhs
            case "<":
                return lhs < rhs
            case ">=":
                return lhs >= rhs
            case "<=":
                return lhs <= rhs
            case "==":
                return lhs == rhs
            case "!=":
                return lhs != rhs
            case "&&":
                return lhs and rhs
            case "||":
                return lhs or rhs
            case _:
                # | ^ & << >> % **
                raise NotImplementedError(f"Unsupported operator '{op}'")

    def visit_UnaryExpression(self, node: ast.UnaryExpression, context: QasmContext):
        ex = self.visit(node.expression, context)
        op = node.op.name
        match op:
            case "-":
                return -ex
            case "~":
                return ~ex
            case "!":
                return not ex
            case _:
                raise NotImplementedError(f"Unsupported operator '{op}'")

    def visit_IntegerLiteral(self, node: ast.IntegerLiteral, context: QasmContext) -> int:
        return node.value

    def visit_Identifier(self, node: ast.Identifier, context: QasmContext):
        id_name = node.name

        if id_name.startswith("$"):
            return self.builder.hw.qubit_with_index(int(id_name.strip("$")))

        qubits = context.registers.quantum.get(id_name, None)
        if qubits is not None:
            return qubits

        bits = context.registers.classic.get(id_name, None)
        if bits is not None:
            return bits

        variable = context.variables.get(id_name, None)
        if variable is not None:
            return variable.value

        # Return constant values
        match id_name:
            case "pi" | "π":
                return math.pi
            case "tau" | "𝜏":
                return math.tau
            case "euler" | "ℇ":
                return math.e

        return id_name

    def visit_IndexedIdentifier(
        self, node: ast.IndexedIdentifier, context: QasmContext
    ) -> list:
        indices = node.indices
        if isinstance(indices, list):
            indices = indices[0]
        frame = self.visit(node.name, context)
        match frame:
            case QubitRegister():
                return [frame.qubits[self.visit(ind, context)] for ind in indices]
            case BitRegister():
                return [frame.bits[self.visit(ind, context)] for ind in indices]
            case _:
                raise NotImplementedError(
                    f"IndexIdentifier not implemented for '{type(frame)}' indexing."
                )

    def visit_QuantumGate(self, node: ast.QuantumGate, context: QasmContext):
        target_qubits = self._get_qubits(node.qubits, context)
        gate_name = node.name.name
        arguments = self._get_node_arguments(node, context)

        match gate_name:
            case "u" | "U":
                theta, phi, _lambda = arguments
                return self.add_unitary(theta, phi, _lambda, target_qubits, self.builder)
            case "cx" | "CX":
                return self.add_cnot(*target_qubits, self.builder)
            case "ecr" | "ECR":
                return self.add_ecr(target_qubits, self.builder)

        if len(node.modifiers) > 0:
            raise NotImplementedError("Gate modifiers are not yet supported.")

        gate_context = QasmContext(registers=Registers(), gates=context.gates)

        if (gate_def := context.gates.get(gate_name, None)) is not None:
            for arg, value in zip(gate_def.arguments, arguments):
                if (known_var := context.variables.get(arg.name, None)) is not None:
                    self._attempt_declaration(
                        Variable(name=known_var.name, var_type=type(value), value=value),
                        gate_context,
                    )
                else:
                    self._attempt_declaration(
                        Variable(
                            name=self.visit(arg, context), var_type=type(value), value=value
                        ),
                        gate_context,
                    )
            for qb_name, value in zip(gate_def.qubits, target_qubits):
                if isinstance(qb_name, (QubitRegister, Qubit)):
                    continue
                self._attempt_declaration(
                    Variable(name=qb_name.name, var_type=type(value), value=value),
                    gate_context,
                )
            for n in gate_def.body:
                self.visit(n, gate_context)

    def visit_QuantumMeasurementStatement(
        self, node: ast.QuantumMeasurementStatement, context: QasmContext
    ):
        qubits = self.visit(node.measure, context)
        bits = self.visit(node.target, context)
        self.add_measure(qubits, bits, self.builder)

    def visit_QuantumMeasurement(self, node: ast.QuantumMeasurement, context: QasmContext):
        return self.visit(node.qubit, context)

    def walk_node(self, node, context: QasmContext, builder, **kwargs):
        pass

    def _assign_returns(self, context: QasmContext):
        register_keys = context.registers.classic.keys()
        if any(register_keys):
            for key in register_keys:
                self.builder.assign(
                    key,
                    [val.value for val in context.registers.classic[key].bits],
                )

            self.builder.returns([key for key in register_keys])

    def add_delay(
        self, delay: float, qubits: list[Qubit], builder: QuantumInstructionBuilder
    ):
        self._add_delay(delay, qubits, builder)

    def add_qreg(
        self,
        reg_name: str,
        reg_length: int,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
    ):
        self._add_qreg(reg_name, reg_length, context, builder)

    def add_creg(self, reg_name: str, reg_length: int, context: QasmContext):
        self._add_creg(reg_name, reg_length, context)

    def add_measure(
        self,
        qubits: list[Qubit],
        bits: list[CregIndexValue],
        builder: QuantumInstructionBuilder,
    ):
        self._add_measure(qubits, bits, builder)

    def add_unitary(
        self,
        theta: float,
        phi: float,
        _lambda: float,
        qubit_or_register: list[Qubit | QubitRegister],
        builder: QuantumInstructionBuilder,
    ):
        self._add_unitary(theta, phi, _lambda, qubit_or_register, builder)

    def add_cnot(
        self,
        control_qbs: list[Qubit],
        target_qbs: list[Qubit],
        builder: QuantumInstructionBuilder,
    ):
        self._add_cnot(control_qbs, target_qbs, builder)

    def add_reset(self, qubits, builder: QuantumInstructionBuilder):
        self._add_reset(qubits, builder)

    def add_if(
        self, left, right, if_body, context: QasmContext, builder: QuantumInstructionBuilder
    ):
        self._add_if(left, right, if_body, context, builder)

    def add_ecr(self, qubits, builder: QuantumInstructionBuilder):
        self._add_ecr(qubits, builder)


def _create_lark_parser():
    with open(
        join(
            Path(__file__).parents[3],
            "frontend",
            "parsers",
            "grammars",
            "partial_qasm3.lark",
        ),
        "r",
        encoding="utf-8",
    ) as lark_grammar_file:
        lark_grammar_str = lark_grammar_file.read()
    return Lark(lark_grammar_str, regex=True)


class Qasm3Parser(Interpreter, AbstractParser):
    lark_parser = _create_lark_parser()

    def __init__(self):
        super().__init__()
        self.builder: QuantumInstructionBuilder | None = None
        self._general_context: OpenPulseContext | None = None
        self._calibration_context: OpenPulseContext | None = None
        self._current_context: OpenPulseContext | None = None
        self._q3_patcher: Qasm3ParserBase = Qasm3ParserBase()
        self._port_mappings: dict[str, PhysicalChannel] = dict()
        self._frame_mappings: dict[str, PulseChannel] = dict()
        self._cached_parses: dict[int, Any] = dict()

        self._has_qasm_version = False
        self._has_calibration_version = False
        self._has_open_pulse = False

    def parser_language(self) -> Languages:
        return Languages.Qasm3

    def _fetch_or_parse(self, qasm_str: str):
        # If we have seen this file before.
        qasm_id = hash(qasm_str)
        if (cached_value := self._cached_parses.get(qasm_id, None)) is not None:
            return cached_value

        try:
            program = self.lark_parser.parse(qasm_str)
        except UnexpectedCharacters as e:
            invalid_string = e._context.strip(" ^\t\n\r")
            raise ValueError(f"Invalid QASM 3 syntax: '{invalid_string}'.")

        self._cached_parses[qasm_id] = program
        return program

    def can_parse(self, qasm_str: str) -> ParseResults:
        try:
            self._fetch_or_parse(qasm_str)
            return ParseResults(success=True)
        except Exception as ex:
            return ParseResults(success=True, errors=str(ex))

    def include(self, tree):
        filename = self.transform_to_value(tree.children[0])
        self._q3_patcher.visit(ast.Include(filename), self._current_context)

    def _reset_and_return(self):
        builder = self.builder
        self.builder = None
        self._general_context = None
        self._calibration_context = None
        self._current_context = None
        self._port_mappings = dict()
        self._frame_mappings = dict()
        self._has_calibration_version = False
        self._has_qasm_version = False
        self._has_open_pulse = False
        return builder

    def walk_node(self, node, context: QasmContext, builder, **kwargs):
        self.visit_children(node)

    def initialize(self, builder: QuantumInstructionBuilder):
        self.builder = builder
        self._q3_patcher.builder = builder

        # Both contexts share global state except for variables.
        self._general_context = self._q3_patcher.load_default_gates(OpenPulseContext())
        self._calibration_context = OpenPulseContext(
            registers=self._general_context.registers,
            gates=self._general_context.gates,
            calibration_methods=self._general_context.calibration_methods,
        )
        self._current_context = self._general_context
        self._frame_mappings = get_frame_mappings(builder.hw)
        self._port_mappings = get_port_mappings(builder.hw)

    def get_waveform_samples(self, waveform: Waveform | SampledWaveform) -> np.ndarray:
        if isinstance(waveform, SampledWaveform):
            return waveform.samples
        else:
            # TODO: how do we do this arbitarily? (COMPILER-752)
            dt = 0.5e-9
            samples = int(waveform.width / dt)
            midway_time = waveform.width / 2

            t = np.linspace(-midway_time, midway_time, samples)
            return waveform.sample(t).samples

    def _perform_signal_processing(
        self, name: str, args
    ) -> Waveform | SampledWaveform | None:
        if name == "mix":
            wf1, wf2 = args
            # TODO: just make getwfsamp take args
            samples1, samples2 = [self.get_waveform_samples(w) for w in (wf1, wf2)]
            output_length, pulse_length = (
                max(len(samples1), len(samples2)),
                min(len(samples1), len(samples2)),
            )
            output = np.append(
                np.array([1] * pulse_length, dtype=complex),
                np.array([0] * (output_length - pulse_length), dtype=complex),
            )
            for wave in (samples1, samples2):
                for i, val in enumerate(wave):
                    output[i] *= val

            return SampledWaveform(samples=output)

        elif name == "sum":
            wf1, wf2 = args
            # TODO: just make getwfsamp take args
            samples1, samples2 = [self.get_waveform_samples(w) for w in (wf1, wf2)]
            pulse_length = max(len(samples1), len(samples2))
            output = np.array([0] * pulse_length, dtype=complex)
            for wave in (samples1, samples2):
                for i, val in enumerate(wave):
                    output[i] += val
            return SampledWaveform(samples=output)

        elif name == "phase_shift":
            wf1: Waveform | SampledWaveform
            wf1, shift = args
            exp_shift = np.exp(1j * shift)
            if isinstance(wf1, SampledWaveform):
                wf1.samples = [exp_shift * val for val in wf1.samples]
            else:
                wf1.phase += shift
            return wf1

        elif name == "scale":
            wf1, scale = args
            if isinstance(wf1, SampledWaveform):
                wf1.samples = [scale * val for val in wf1.samples]
            else:
                wf1.scale_factor *= scale
            return wf1

        return None

    def _attempt_declaration(self, var: Variable):
        if var.name in self._current_context.variables:
            raise ValueError(f"Can't redeclare variable {var.name}")
        self._current_context.variables[var.name] = var

    def transform_to_value(self, child_tree, walk_variable=True, return_variable=False):
        if isinstance(child_tree, list):
            return [
                self.transform_to_value(val, walk_variable, return_variable)
                for val in child_tree
            ]

        if not isinstance(child_tree, (Tree, Token)):
            return child_tree

        # TODO: For now it's fine, but want to change all visitor methods to return
        #   appropriate values and chain from there, as right now arguments and root
        #   calls are treated differently.
        def _walk_to_single_value(node: Tree):
            if not isinstance(node, Tree):
                return node, None

            if (
                any([isinstance(node, Token) for node in node.children])
                or len(node.children) > 1
            ):
                if len(node.children) == 1:
                    return node.children[0], node.data
                return node, node.data
            else:
                return _walk_to_single_value(node.children[0])

        node, data = _walk_to_single_value(child_tree)
        if isinstance(node, Tree):
            if data == "real_number":
                op, value = self.transform_to_value(node.children)
                if op == "-":
                    return -value
                elif op == "+":
                    return +value
            elif data == "unary_expression":
                op, value = self.transform_to_value(node.children)
                if op == "-":
                    return -value
                elif op == "~":
                    return ~value
                elif op == "!":
                    return not value
            elif data == "index_identifier":
                registers = self.transform_to_value(node.children[0])
                if isinstance(registers, BitRegister):
                    reg_index = self.transform_to_value(node.children[1])
                    return next(
                        (val for val in registers.bits if val.index == reg_index), None
                    )

                if isinstance(registers, QubitRegister):
                    reg_index = self.transform_to_value(node.children[1])
                    if len(qubits := registers.qubits) > reg_index:
                        return qubits[reg_index]
                    return None

            if data == "additive_expression":
                return self.transform_to_value(node.children[0]) + self.transform_to_value(
                    node.children[2]
                )

            elif data == "subtraction_expression":
                return self.transform_to_value(node.children[0]) - self.transform_to_value(
                    node.children[2]
                )

            elif data == "division_expression":
                return self.transform_to_value(node.children[0]) / self.transform_to_value(
                    node.children[2]
                )

            elif data == "multiplicative_expression":
                return self.transform_to_value(node.children[0]) * self.transform_to_value(
                    node.children[2]
                )

            elif data == "complex_number":
                return complex(
                    "".join([str(self.transform_to_value(val)) for val in node.children])
                )

            elif data == "timing_literal":
                value = self.transform_to_value(node.children[0])
                unit = self.transform_to_value(node.children[1])
                # TODO: [JF] Work out what DT is meant to be here.
                if unit == "dt":
                    # DT hardcoded at .5ns
                    return value / 2000000000
                elif unit == "ns":
                    return value / 1000000000
                elif unit == "us" or unit == "µs":
                    return value / 1000000
                elif unit == "ms":
                    return value / 1000
                elif unit == "s":
                    return value

            elif data == "extern_or_subroutine_call":
                return self.visit(node)

            # Return tuple of 'type' and 'name' if we're a classical argument.
            elif data == "classical_argument":
                results = [
                    self.transform_to_value(val, walk_variable, return_variable)
                    for val in node.children
                ]
                if len(results) == 1:
                    return None, results[0]

                return tuple(results)

            return [
                self.transform_to_value(val, walk_variable, return_variable)
                for val in node.children
            ]

        node: Token
        if data == "imag_number":
            return complex(0, float(node.value))
        elif node.type == "FLOAT_LITERAL":
            return float(node.value)
        elif node.type == "DECIMAL_INTEGER_LITERAL":
            return int(node.value)
        elif node.type == "CONSTANT":
            if node.value in ("pi", "π"):
                return math.pi
            elif node.value in ("tau", "𝜏"):
                return math.tau
            elif node.value in ("euler", "ℇ"):
                return math.e
        elif node.type == "IDENTIFIER":
            # $num just means get qubit at that index
            id_value: str = node.value
            if id_value.startswith("$"):
                return self.builder.hw.qubit_with_index(int(id_value.strip("$")))

            qubits = self._current_context.registers.quantum.get(id_value, None)
            if qubits is not None:
                return qubits

            bits = self._current_context.registers.classic.get(id_value, None)
            if bits is not None:
                return bits

            def _walk_variables(var_id, guard: set[str] = None):
                if guard is None:
                    guard = set()
                variable = self._current_context.variables.get(var_id, None)
                if variable is not None:
                    if walk_variable:
                        if isinstance(variable.value, Variable):
                            if var_id not in guard:
                                guard.add(var_id)
                                return _walk_variables(variable.name, guard)

                        return variable if return_variable else variable.value
                return var_id

            return _walk_variables(id_value)

        # If we have encompassing quotes just strip them as we're already a string.
        node_value = node.value
        if node_value.count('"') == 2:
            node_value = node_value.lstrip('"')
            node_value = node_value.rstrip('"')

        return node_value

    def version(self, tree: Tree):
        version_number = self.transform_to_value(tree)
        if version_number != 3:
            raise ValueError("OpenQASM version has to be 3.")
        self._has_qasm_version = True

    def calibration_grammar_declaration(self, tree: Tree):
        version_number = self.transform_to_value(tree)[1]
        if version_number != "openpulse":
            raise ValueError("Calibration grammar has to be openpulse.")
        self._has_calibration_version = True

    def frame_definition(self, tree: Tree):
        name = self.transform_to_value(tree.children[1])
        args = self.transform_to_value(tree.children[4])
        port = args[0]
        frequency = (
            self._get_frequency(args[1][0]) if isinstance(args[1], list) else args[1]
        )
        phase = 0.0 if len(args) <= 2 else args[2]

        if isinstance(port, IRPulseChannel):
            physical_channel_id = port.physical_channel_id
        elif isinstance(port, PhysicalChannel):
            physical_channel_id = port.uuid
        else:
            raise TypeError(
                f"Cannot create new frame from variable '{name}'. "
                "Must be either type Port or Frame."
            )

        pulse_channel = self.builder.create_pulse_channel(
            frequency=frequency, physical_channel=physical_channel_id
        )
        self.builder.phase_shift(target=pulse_channel, theta=phase)

        self._attempt_declaration(
            Variable(name=name, var_type=IRPulseChannel, value=pulse_channel)
        )

    def waveform_definition(self, tree: Tree):
        if len(tree.children) == 4:
            assigned_variable = self.transform_to_value(tree.children[1])
            intrinsic_name = None
        elif len(tree.children) == 5:
            assigned_variable = self.transform_to_value(tree.children[1])
            intrinsic_name = self.transform_to_value(tree.children[3])
        else:
            raise ValueError("Unknown waveform definition.")

        _empty = tuple()

        def _validate_waveform_args(
            *,
            width=_empty,
            amp=_empty,
            beta=_empty,
            zero_at_edges=_empty,
            rise=_empty,
            std_dev=_empty,
            frequency=_empty,
            phase=_empty,
            square_width=_empty,
            drag=_empty,
        ):
            if width is not _empty and not isinstance(width, float):
                raise ValueError(
                    f"Width '{str(width)}' used in {intrinsic_name} is not a float."
                )
            if drag is not _empty and not isinstance(drag, float):
                raise ValueError(
                    f"Drag '{str(width)}' used in {intrinsic_name} is not a float."
                )
            if amp is not _empty and not isinstance(amp, Number):
                raise ValueError(
                    f"Amp '{str(amp)}' used in {intrinsic_name} is not a float."
                )
            if beta is not _empty and not isinstance(beta, float):
                raise ValueError(
                    f"Beta '{str(beta)}' used in {intrinsic_name} is not a float."
                )
            if zero_at_edges is not _empty and not isinstance(zero_at_edges, int):
                raise ValueError(
                    f"Zero at edges '{str(zero_at_edges)}' used in {intrinsic_name} "
                    "is not a int."
                )
            if rise is not _empty and not isinstance(rise, float):
                raise ValueError(
                    f"Rise '{str(rise)}' used in {intrinsic_name} is not a float."
                )
            if std_dev is not _empty and not isinstance(std_dev, (float, int)):
                raise ValueError(
                    f"Standard deviation '{str(std_dev)}' used in {intrinsic_name} "
                    "is not a float."
                )
            if frequency is not _empty and not isinstance(frequency, float):
                raise ValueError(
                    f"Frequency '{str(frequency)}' used in {intrinsic_name} is not a float."
                )
            if phase is not _empty and not isinstance(phase, float):
                raise ValueError(
                    f"Phase '{str(phase)}' used in {intrinsic_name} is not a float."
                )
            if square_width is not _empty and not isinstance(square_width, float):
                raise ValueError(
                    f"Square width '{str(square_width)}' used in {intrinsic_name} "
                    "is not a float."
                )

        def _validate_arg_length(arg_tree, *lengths):
            """As results length is dynamic, centralise validation and messages."""
            lengths = set(lengths)
            args = self.transform_to_value(arg_tree)
            is_iterable = isinstance(args, (tuple, list))
            arg_length = len(args) if is_iterable else 1
            if (is_iterable and arg_length not in lengths) or (
                not is_iterable and 1 not in lengths
            ):
                raise ValueError(
                    f"Waveform '{intrinsic_name}' has incorrect number of arguments. "
                    f"Needs {','.join([str(val) for val in lengths])}, "
                    f"has {arg_length}."
                )
            return args + ([None] * (max(lengths) - arg_length))

        match intrinsic_name:
            case None:
                # This is a flat array of pulse values.
                array_contents = self.transform_to_value(tree.children[3])
                waveform = SampledWaveform(samples=array_contents)

            # TODO: implement non intrinsic waveforms.
            case "constant":
                width, amp = _validate_arg_length(tree.children[4], 2)
                _validate_waveform_args(width=width, amp=amp)
                waveform = SquareWaveform(width=width, amp=amp)

            case "rounded_square":
                width, std_dev, rise_time, amp = _validate_arg_length(tree.children[4], 4)
                _validate_waveform_args(
                    width=width, std_dev=std_dev, amp=amp, rise=rise_time
                )
                waveform = RoundedSquareWaveform(
                    width=width,
                    std_dev=std_dev,
                    amp=amp,
                    rise=rise_time,
                )

            case "drag":
                amp, width, std_dev, beta, zero_at_edges = _validate_arg_length(
                    tree.children[4], 4, 5
                )
                zero_at_edges = 0 if not zero_at_edges else 1
                _validate_waveform_args(
                    width=width,
                    amp=amp,
                    beta=beta,
                    zero_at_edges=zero_at_edges,
                    std_dev=std_dev,
                )
                waveform = DragGaussianWaveform(
                    width=width,
                    amp=amp,
                    zero_at_edges=zero_at_edges,
                    beta=beta,
                    std_dev=std_dev,
                )

            case "gaussian":
                amp, width, std_dev = _validate_arg_length(tree.children[4], 3)
                _validate_waveform_args(width=width, amp=amp, std_dev=std_dev)
                waveform = GaussianZeroEdgeWaveform(
                    width=width,
                    amp=amp,
                    std_dev=std_dev,
                    zero_at_edges=0,
                )

            case "gaussian_zero_edge":
                amp, width, std_dev, zero_at_edges = _validate_arg_length(
                    tree.children[4], 4
                )
                zero_at_edges = bool(zero_at_edges)
                _validate_waveform_args(
                    width=width, amp=amp, zero_at_edges=zero_at_edges, std_dev=std_dev
                )
                waveform = GaussianZeroEdgeWaveform(
                    width=width,
                    amp=amp,
                    zero_at_edges=zero_at_edges,
                    std_dev=std_dev,
                )

            case "sech":
                amp, width, std_dev = _validate_arg_length(tree.children[4], 3)
                waveform = SechWaveform(width=width, amp=amp, std_dev=std_dev)

            case "gaussian_square":
                amp, width, square_width, std_dev, zero_at_edges = _validate_arg_length(
                    tree.children[4], 4, 5
                )
                zero_at_edges = bool(zero_at_edges)
                _validate_waveform_args(
                    width=width,
                    amp=amp,
                    square_width=square_width,
                    std_dev=std_dev,
                    zero_at_edges=zero_at_edges,
                )
                waveform = GaussianSquareWaveform(
                    width=width,
                    std_dev=std_dev,
                    amp=amp,
                    zero_at_edges=zero_at_edges,
                    square_width=square_width,
                )

            case "sine":
                amp, width, frequency, phase = _validate_arg_length(tree.children[4], 4)
                _validate_waveform_args(
                    width=width, amp=amp, frequency=frequency, phase=phase
                )
                waveform = SinWaveform(
                    amp=amp,
                    width=width,
                    frequency=frequency,
                    phase=phase,
                )

            case "gaussian_rise":
                amp, width, rise, drag, phase = _validate_arg_length(tree.children[4], 5)
                _validate_waveform_args(
                    width=width, rise=rise, amp=amp, drag=drag, phase=phase
                )
                waveform = GaussianWaveform(
                    amp=amp,
                    width=width,
                    rise=rise,
                    drag=drag,
                    phase=phase,
                )

            case "soft_square_rise":
                amp, width, rise, drag, phase = _validate_arg_length(tree.children[4], 5)
                _validate_waveform_args(
                    width=width, rise=rise, amp=amp, drag=drag, phase=phase
                )
                waveform = SoftSquareWaveform(
                    amp=amp,
                    width=width,
                    rise=rise,
                    drag=drag,
                    phase=phase,
                )

            case _:

                def snake_case_to_CamelCase(input: str):
                    return input.replace("_", " ").title().replace(" ", "")

                waveform_type = locate(
                    "qat.ir.waveforms."
                    + snake_case_to_CamelCase(intrinsic_name)
                    + "Waveform"
                )
                if waveform_type is None:
                    raise ValueError(f"Unknown waveform {intrinsic_name}.")

                width, amp = _validate_arg_length(tree.children[4], 2)
                _validate_waveform_args(width=width, amp=amp)
                waveform = waveform_type(width=width, amp=amp)

        self._attempt_declaration(
            Variable(name=assigned_variable, var_type=AbstractWaveform, value=waveform)
        )

    def timing_instruction(self, tree: Tree):
        """This is actually a delay instruction."""
        delay_time, target = self.transform_to_value(tree)
        self.builder.delay(target, delay_time)

    def quantum_measurement_assignment_statement(self, tree: Tree):
        args = self.transform_to_value(tree)
        if len(args) == 2:
            qubits = args[0]
            bits = args[1]
        else:
            bits = args[0]
            qubits = args[2]

        qubits = qubits if isinstance(qubits, list) else [qubits]

        if isinstance(bits, BitRegister):
            bits = bits.bits
        bits = bits if isinstance(bits, list | ValidatedList) else [bits]

        # If the measure for this particular qubit has been overriden the functionality
        # is distinctly different.
        if self._has_defcal_override("measure", qubits):
            results = self._call_gate("measure", qubits)
            if results is not None:
                results = results if isinstance(results, list) else [results]
                if len(bits) != len(results):
                    raise ValueError("Can't flatten overriden measure into assignment.")

                bit: CregIndexValue
                for bit, result in zip(bits, results):
                    bit.value = result
        else:
            self._q3_patcher.add_measure(qubits, bits, self.builder)

    def _has_defcal_override(
        self,
        name: str,
        qubits: list[Qubit],
        argument_values: list[Any] | None = None,
    ):
        """
        Returns whether this gate has been overriden, either in a generic
        or qubit-specific manner.
        """
        argument_values = argument_values or []
        qubits = qubits if isinstance(qubits, list) else [qubits]
        qubit_specific_name = self._create_qb_specific_gate_suffix(name, qubits)

        is_calibration = name in self._current_context.calibration_methods
        is_qubit_specific = qubit_specific_name in self._current_context.calibration_methods
        is_list_expr = (
            self.generate_expr_list_defcal_name(name, argument_values)
            in self._current_context.calibration_methods
        )

        return is_calibration or is_qubit_specific or is_list_expr

    def generate_expr_list_defcal_name(self, name, expr_list):
        if not isinstance(expr_list, list):
            expr_list = [expr_list]
        return name + "_" + "_".join([str(i) for i in expr_list])

    def _call_gate(self, name, method_args, throw_on_missing=True):
        qubits, others = [], []

        def _strip_qubits(value):
            if isinstance(value, list):
                for val in value:
                    _strip_qubits(val)
            else:
                if isinstance(value, (Qubit, QubitRegister)):
                    qubits.append(value)
                else:
                    others.append(value)

        _strip_qubits(method_args)

        gate_def = self._current_context.gates.get(name, None)
        cal_def = self._current_context.calibration_methods.get(name, None)
        new_name = self.generate_expr_list_defcal_name(name, others)
        expr_list = False
        cal_def_new = self._current_context.calibration_methods.get(new_name, None)
        cal_def = cal_def_new or cal_def
        if cal_def_new is not None:
            expr_list = True

        qb_specific_cal_def = self._current_context.calibration_methods.get(
            self._create_qb_specific_gate_suffix(name, qubits), None
        )

        qb_specific_cal_def_expr_list = self._current_context.calibration_methods.get(
            self._create_qb_specific_gate_suffix(new_name, qubits), None
        )

        # Functionally the methods are processed the same, but the qb-specific override
        # takes presidence.
        if qb_specific_cal_def is not None:
            cal_def = qb_specific_cal_def

        if qb_specific_cal_def_expr_list is not None:
            expr_list = True
            cal_def = qb_specific_cal_def_expr_list

        if name in ("u", "U"):
            # u is not in openpulse grammar so cannot be overridden...
            self._q3_patcher.add_unitary(
                others[0], others[1], others[2], qubits, self.builder
            )

        # Prioritize calibration definitions here if people override the base functions.
        # We also don't care about qubit scoping and restrictions.
        elif cal_def is not None:
            # Implied 'barrier' between defcals. To make things simple just assume that
            # everything in the defcal focuses on the qubits coming in.
            self.builder.synchronize(qubits)
            existing_context = self._current_context
            self._current_context = OpenPulseContext(
                registers=self._general_context.registers,
                gates=self._general_context.gates,
                variables=dict(self._calibration_context.variables),
                calibration_methods=self._general_context.calibration_methods,
            )

            arg_mappings, qubit_mappings, body = cal_def
            if not expr_list:
                # Strip off the type as we have no need for it here.
                # No type given if an expression list
                arg_mappings = [name for _, name in arg_mappings]

            if not isinstance(arg_mappings, list):
                arg_mappings = [arg_mappings]

            if not isinstance(qubit_mappings, list):
                qubit_mappings = [qubit_mappings]

            if len(arg_mappings) != len(others):
                raise ValueError(
                    f"Call to '{name}' needs {len(arg_mappings)} arguments. "
                    f"Has {len(others)}."
                )

            for arg_name, value in zip(arg_mappings, others):
                self._attempt_declaration(
                    Variable(name=str(arg_name), var_type=type(value), value=value)
                )

            for qb_name, value in zip(qubit_mappings, qubits):
                # If we resolved to a qubit already, we're a physical qubit.
                if isinstance(qb_name, (QubitRegister, Qubit)):
                    continue

                self._attempt_declaration(
                    Variable(name=qb_name, var_type=type(value), value=value)
                )

            cali_results = self.visit_children(body)
            self._current_context = existing_context

            # We imply that the final returned value is the 'return' from the
            # calibration block.
            # If people want to return something distinct, use a return.
            ret_value = cali_results[-1]
            return ret_value

        elif gate_def is not None:
            # Our wrapper exposes any fields required and we override the argument
            # gathering.
            node = ast.QuantumGate(
                name=ast.Identifier(name), qubits=qubits, arguments=others, modifiers=[]
            )
            self._q3_patcher.visit(node, self._current_context)
        elif name in ("cnot", "CNOT"):
            self._q3_patcher.add_cnot(qubits[0], qubits[1], self.builder)
        elif name in ("ecr", "ECR"):
            self._q3_patcher.add_ecr(qubits, self.builder)
        elif throw_on_missing:
            raise ValueError(
                f"Can't find gate implementation for '{name}' with supplied arguments."
            )

    def quantum_gate_call(self, tree: Tree):
        args = self.transform_to_value(tree)
        name = args[0]
        method_args = args[1:]
        self._call_gate(name, method_args)

    def quantum_reset(self, tree: Tree):
        targets = self.transform_to_value(tree)
        self.builder.reset(targets)

    def quantum_barrier(self, tree: Tree):
        args = self.transform_to_value(tree)
        self.builder.synchronize(args)

    def quantum_measurement(self, tree: Tree):
        """Throwaway measurement that doesn't store results into any variable."""
        qubits = self.transform_to_value(tree)

        if self._has_defcal_override("measure", qubits):
            self._call_gate("measure", qubits)
        else:
            self.builder.measure_single_shot_z(qubits)

    def return_statement(self, tree: Tree):
        variable = self.transform_to_value(tree, walk_variable=False)

        # Expressions return variables directly, so no need for lookup.
        if not isinstance(variable, Variable):
            if isinstance(variable, str):
                var_name = variable
                variable = self._current_context.variables.get(var_name, None)
                if variable is None:
                    raise ValueError(f"Can't return {var_name} as it doesn't exist.")
            else:
                variable = Variable.with_random_name()
                variable.value = variable
                self._attempt_declaration(variable)

        return variable

    # Timed box, not used right now.
    def timing_box(self, tree: Tree):
        pass

    def quantum_declaration(self, tree: Tree):
        args = self.transform_to_value(tree)
        if isinstance(args, str):
            length = 1
            variable = args
        else:
            # If you're using the soon-to-be-depreciated qreg,
            # arguments are switched.
            if isinstance(args[0], str):
                length = args[1]
                variable = args[0]
            else:
                length = args[0]
                variable = str(args[1])

        self._q3_patcher.add_qreg(variable, length, self._current_context, self.builder)

    def bit_declaration_statement(self, tree: Tree):
        args = self.transform_to_value(tree)
        if isinstance(args, str):
            length = 1
            variable = args
        else:
            # If you are using the soon-to-be-depreciated creg,
            # arguments are switched.
            if isinstance(args[0], str):
                length = args[1]
                variable = args[0]
            else:
                length = args[0]
                variable = args[1]

        self._q3_patcher.add_creg(variable, length, self._current_context)

    def complex_declaration_statement(self, tree: Tree):
        _, name = self.transform_to_value(tree)

        self._attempt_declaration(Variable(name=name, var_type=complex))

    def _create_qb_specific_gate_suffix(self, name: str, target_qubits: list[Qubit]):
        return f"{name}[{','.join([str(qb) for qb in target_qubits])}]"

    def calibration_definition(self, tree: Tree):
        self._has_open_pulse = True
        gate_name: str = self.transform_to_value(tree.children[1])
        is_expr_list = False
        if len(tree.children) == 4:
            classic_args = []
            target_qubits = self.transform_to_value(tree.children[2], walk_variable=False)
            body = tree.children[3]
        else:
            is_expr_list = tree.children[2].children[0].data == "expression_list"
            classic_args = self.transform_to_value(tree.children[2], walk_variable=False)
            target_qubits = self.transform_to_value(tree.children[3], walk_variable=False)
            body = tree.children[4]

        target_qubits = (
            target_qubits if isinstance(target_qubits, list) else [target_qubits]
        )
        classic_args = classic_args if isinstance(classic_args, list) else [classic_args]

        if is_expr_list:
            gate_name = self.generate_expr_list_defcal_name(gate_name, classic_args)

        if all(isinstance(val, (Qubit, QubitRegister)) for val in target_qubits):
            gate_name = self._create_qb_specific_gate_suffix(gate_name, target_qubits)
        self._current_context.calibration_methods[gate_name] = (
            classic_args,
            target_qubits,
            body,
        )

    def gate_definition(self, tree: Tree):
        gate_name: str = self.transform_to_value(tree.children[0])
        if len(tree.children) == 3:
            classic_args = []
            target_qubits = self.transform_to_value(tree.children[1], walk_variable=False)
            body = tree.children[2]
        else:
            classic_args = self.transform_to_value(tree.children[1], walk_variable=False)
            target_qubits = self.transform_to_value(tree.children[2], walk_variable=False)
            body = tree.children[3]

        target_qubits = (
            target_qubits if isinstance(target_qubits, list) else [target_qubits]
        )
        if all(isinstance(val, (Qubit, QubitRegister)) for val in target_qubits):
            gate_name = self._create_qb_specific_gate_suffix(gate_name, target_qubits)

        # Not technically a calibration method, but the way to call is the same.
        self._current_context.calibration_methods[gate_name] = (
            classic_args,
            target_qubits,
            body,
        )

    def _get_phase(self, pulse_channel: IRPulseChannel):
        phase = 0
        for inst in self.builder.instructions:
            inst: QuantumInstruction
            if isinstance(inst, PhaseShift) and pulse_channel.uuid in inst.targets:
                phase += inst.phase
            elif isinstance(inst, PhaseReset) and pulse_channel.uuid in inst.targets:
                phase = 0
        return phase

    def _get_frequency(self, pulse_channel: IRPulseChannel):
        frequency = pulse_channel.frequency
        for inst in self.builder.instructions:
            if isinstance(inst, FrequencyShift) and pulse_channel.uuid in inst.targets:
                frequency += inst.frequency
        return frequency

    def _capture_iq_value(
        self,
        pulse_channel: IRPulseChannel,
        duration: float,
        output_variable: str,
        filter: Pulse = None,
    ):
        # The acquire integrator mode means that for every acquired resonator response
        # signal within a group of shots, we integrate along the time axis to find the
        # average amplitude of the response.
        #
        # The returned value for each shot after postprocessing is a complex iq value.
        qubit = self.builder.hw.qubit_for_physical_channel_id(
            pulse_channel.physical_channel_id
        )
        if qubit is None:
            raise TypeError(
                f"Pulse channel {pulse_channel} is not assigned to any known qubit."
            )

        # This kind of logic to add delays is really part of a measure definition, and
        # not within the semantics of a capture command, which is essentially an acqurire.
        # I don't think this is necessary to do, but I won't change it for now...
        hw_pulse_channel = self.builder.hw.pulse_channel_with_id(pulse_channel.uuid)
        if isinstance(hw_pulse_channel, AcquirePulseChannel):
            delay = hw_pulse_channel.acquire.delay
        else:
            log.warning(
                f"The acquire channel {pulse_channel.uuid} is not an acquire channel: "
                "setting the delay to 0.0."
            )
            delay = 0.0

        acquire = Acquire(
            targets=pulse_channel.uuid,
            duration=duration,
            mode=AcquireMode.INTEGRATOR,
            output_variable=output_variable,
            filter=filter,
            delay=delay,
        )
        self.builder.add(acquire)
        self.builder.post_processing(
            target=qubit,
            process_type=PostProcessType.MEAN,
            axes=[ProcessAxis.TIME],
            output_variable=output_variable,
        )

        return acquire

    def _validate_channel_args(
        self, pulse_channel: IRPulseChannel, val_type: str, value=None
    ):
        if not isinstance(pulse_channel, IRPulseChannel):
            raise ValueError(f"{str(pulse_channel)} is not a valid pulse channel.")

        if value is not None and not isinstance(value, (int, float)):
            raise ValueError(f"{str(value)} is not a valid {val_type}.")

        return pulse_channel, value

    def _validate_phase_args(self, pulse_channel: IRPulseChannel, phase=None):
        return self._validate_channel_args(pulse_channel, "phase", phase)

    def _validate_freq_args(self, pulse_channel: IRPulseChannel, frequency=None):
        return self._validate_channel_args(pulse_channel, "frequency", frequency)

    def extern_or_subroutine_call(self, tree: Tree):
        name = self.transform_to_value(tree.children[0])
        args = self.transform_to_value(tree.children[1])

        if name in ("mix", "sum", "phase_shift", "scale"):
            return self._perform_signal_processing(name, args)

        elif name == "play":
            waveform: Waveform | SampledWaveform = args[1]
            if not isinstance(waveform, Waveform | SampledWaveform):
                variable_name = self.transform_to_value(
                    tree.children[1].children[1], walk_variable=False
                )
                raise ValueError(
                    f"Play frame argument {variable_name} has not been linked to a frame."
                )

            pulse_target = args[0]
            if not isinstance(pulse_target, IRPulseChannel):
                variable_name = self.transform_to_value(tree.children[1].children[0])
                raise ValueError(
                    f"Play waveform argument {variable_name} does not point to a waveform."
                )

            waveform_blob = deepcopy(waveform)
            pulse = Pulse(
                targets=pulse_target.uuid, waveform=waveform_blob, ignore_channel_scale=True
            )

            self.builder.add(pulse)

        elif name == "shift_phase":
            pulse_channel, phase = self._validate_phase_args(args[0], args[1])
            self.builder.add(PhaseShift(targets=pulse_channel.uuid, phase=phase))

        elif name == "set_phase":
            pulse_channel, phase_arg = self._validate_phase_args(args[0], args[1])
            phase = self._get_phase(pulse_channel)
            self.builder.add(
                PhaseShift(targets=pulse_channel.uuid, phase=phase_arg - phase)
            )

        elif name == "get_phase":
            pulse_channel, _ = self._validate_phase_args(args)
            return self._get_phase(pulse_channel)

        elif name == "set_frequency":
            pulse_channel, freq_arg = self._validate_freq_args(args[0], args[1])
            frequency = self._get_frequency(pulse_channel)
            self.builder.add(
                FrequencyShift(targets=pulse_channel.uuid, frequency=freq_arg - frequency)
            )

        elif name == "get_frequency":
            pulse_channel, _ = self._validate_freq_args(args)
            return self._get_frequency(pulse_channel)

        elif name == "shift_frequency":
            pulse_channel, freq_arg = self._validate_freq_args(args[0], args[1])
            self.builder.add(FrequencyShift(targets=pulse_channel.uuid, frequency=freq_arg))

        elif name == "capture_v0":
            # Not sure what this method should return.
            raise NotImplementedError(
                "capture_v0 is not yet implemented. Please use capture_v1, capture_v2 or "
                "capture_v3 instead."
            )

        elif name == "capture_v1":
            # A capture command that returns an iq value
            variable: Variable = Variable.with_random_name()
            self._capture_iq_value(
                args[0], args[1], variable.name, args[2] if len(args) > 2 else None
            )

            self._attempt_declaration(variable)
            return variable

        elif name == "capture_v2":
            # A capture command that returns a discriminated bit
            # The first part of this capture is the same as capture_v1 but we take the
            # complex iq value and perform a linear complex to real map which is used to
            # adjust the iq values to a form which can be discriminated into a bit. If
            # the value is positive we return 0, qubit is in ground state, if the value
            # is negative we return 1, qubit is in excited state.
            pulse_channel = args[0]
            variable: Variable = Variable.with_random_name()
            acquire = self._capture_iq_value(
                pulse_channel,
                args[1],
                variable.name,
                args[2] if len(args) > 2 else None,
            )  # TODO: Fill in output_variable
            mean_z_map_args = None
            if len(args) > 3:
                mean_z_map_args = args[3]
            else:
                for q in self.builder.hw.qubits.values():
                    if q.resonator.pulse_channels.pulse_channel_with_id(pulse_channel.uuid):
                        mean_z_map_args = q.mean_z_map_args
                        break
            if mean_z_map_args is None:
                keys = next(
                    key
                    for key, value in self._frame_mappings.items()
                    if value == pulse_channel
                )
                raise ValueError(f"Could not find mean_z_map_args for frame {keys}.")
            self.builder.post_processing(
                target=q,
                process_type=PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL,
                args=mean_z_map_args,
                output_variable=acquire.output_variable,
            )
            self.builder.results_processing(
                variable=variable.name, res_format=InlineResultsProcessing.Program
            )

            self._attempt_declaration(variable)
            return variable

        elif name == "capture_v3":
            # A capture command that returns a raw waveform data
            #
            # The acquire scope mode will average the resonator response signal over all
            # the shots within a group.
            #
            # Unlike the integrator mode, these will not be integrated over time to get
            # an average amplitude but will instead keep the entire waveform, averaging
            # over all the shots. This means you might get unexpected results for
            # greater than 1 shots if you are measuring a qubit in superposition, as the
            # different signal responses from the different measurement results will
            # average out.
            variable: Variable = Variable.with_random_name()
            pulse_channel = args[0]

            acquire = Acquire(
                targets=pulse_channel.uuid,
                duration=args[1],
                mode=AcquireMode.SCOPE,
                output_variable=variable.name,
                filter=args[2] if len(args) > 2 else None,
            )
            qubit = None
            for q in self.builder.hw.qubits.values():
                if q.resonator.pulse_channels.pulse_channel_with_id(pulse_channel.uuid):
                    qubit = q
            if not isinstance(qubit, Qubit):
                raise ValueError(
                    f"Pulse channel with id {pulse_channel.uuid} is not associated with any quantum device."
                )

            self.builder.add(acquire)
            self.builder.post_processing(
                target=qubit,
                process_type=PostProcessType.MEAN,
                axes=[ProcessAxis.SEQUENCE],
                output_variable=acquire.output_variable,
            )
            self._attempt_declaration(variable)
            return variable

        elif name == "capture_v4":
            raise NotImplementedError(
                "capture_v4 is not implemented. Please use capture_v1, capture_v2 or "
                "capture_v3 instead."
            )
        else:
            raise ValueError(f"Extern {name} not implemented.")

    def extern_frame(self, tree: Tree):
        name = self.transform_to_value(tree, walk_variable=False)
        hwm_pulse_channel = self._frame_mappings.get(name, None)
        if hwm_pulse_channel is None:
            raise ValueError(f"Could not find extern Frame with name '{name}'.")
        pulse_channel = self.builder.get_pulse_channel(hwm_pulse_channel.uuid)

        self._attempt_declaration(
            Variable(name=name, var_type=IRPulseChannel, value=pulse_channel)
        )

    def extern_port(self, tree: Tree):
        name = self.transform_to_value(tree, walk_variable=False)
        physical_channel = self._port_mappings.get(name, None)
        if physical_channel is None:
            raise ValueError(f"Could not find extern Port with name '{name}'.")

        self._attempt_declaration(
            Variable(name=name, var_type=PhysicalChannel, value=physical_channel)
        )

    def frame_attribute_assignment(self, tree: Tree):
        args = self.transform_to_value(tree)
        pulse_channel = args[0][0]
        if not isinstance(pulse_channel, IRPulseChannel):
            raise ValueError("Tried to assign to a frame that doesn't exist.")

        field = args[0][1]
        op = args[1]
        value = args[2]
        if field == "phase":
            inst = PhaseShift
            getter = self._get_phase
        elif field == "frequency":
            inst = FrequencyShift
            getter = self._get_frequency
        else:
            raise ValueError(f"Attempted to assign to an unknown frame field '{field}'.")

        if op == "=":
            current_value = getter(pulse_channel)
            instr_info = {"targets": pulse_channel.uuid, field: value - current_value}
        elif op == "+=":
            instr_info = {"targets": pulse_channel.uuid, field: value}
        elif op == "-=":
            instr_info = {"targets": pulse_channel.uuid, field: -value}
        else:
            raise ValueError(f"Attempted to use an unknown frame operator '{op}'.")
        self.builder.add(inst(**instr_info))

    def assignment(self, tree: Tree):
        register = self.transform_to_value(tree.children[0], walk_variable=False)
        op = self.transform_to_value(tree.children[1])
        value = self.transform_to_value(tree.children[2])

        if op != "=":
            raise ValueError(f"Assignment operator {op} is unsupported.")

        if isinstance(register, CregIndexValue):
            register = register.register_name

        existing = self._current_context.variables.get(register, None)
        if existing is not None:
            existing.value = value
        else:
            self._attempt_declaration(
                value
                if isinstance(value, Variable)
                else Variable(name=register, var_type=type(value), value=value)
            )

    def cal_block(self, tree: Tree):
        self._has_open_pulse = True
        existing_context = self._current_context
        self._current_context = self._calibration_context
        self.visit_children(tree)
        self._current_context = existing_context

    def parse(self, builder: QuantumInstructionBuilder, qasm_str: str):
        self.initialize(builder)
        parsed = self._fetch_or_parse(qasm_str)

        if (qasm_id := hash(qasm_str)) in self._cached_parses:
            del self._cached_parses[qasm_id]

        # If we have a new results format, propagate it.
        self._q3_patcher.results_format = self.results_format

        self.visit(parsed)
        if not self._has_qasm_version:
            raise ValueError("Ambiguous QASM version, need OPENQASM header.")

        if self._has_open_pulse and not self._has_calibration_version:
            raise ValueError("Uses pulse definitions without defcalgrammar header.")

        # If we're purely QASM act like the previous version in regards to results.
        register_keys = self._general_context.registers.classic.keys()
        if any(register_keys):
            for key in register_keys:
                builder.assign(
                    key,
                    [
                        val.value
                        for val in self._general_context.registers.classic[key].bits
                    ],
                )

            builder.returns([key for key in register_keys])

        return self._reset_and_return()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from qat.model.hardware_model import PydLogicalHardwareModel, PydPhysicalHardwareModel
from qat.purr.compiler.hardware_models import LegacyHardwareModel

LegHwModel = TypeVar("LegHwModel", bound=LegacyHardwareModel)

PydHwModel = TypeVar("PydHwModel", PydPhysicalHardwareModel, PydLogicalHardwareModel)

HwModel = TypeVar(
    "HwModel", LegacyHardwareModel, PydLogicalHardwareModel, PydPhysicalHardwareModel
)


class BaseModelLoader(Generic[HwModel], ABC):
    """ModelLoaders load HardwareModels from a source configured on initialisation."""

    def __init__(self): ...

    @abstractmethod
    def load(self) -> HwModel:
        """Load and return the Hardware Model.

        :return: A loaded Hardware Model
        :rtype: LegacyHardwareModel | LogicalHardwareModel
        """
        pass


class BaseLogicalModelLoader(BaseModelLoader[PydLogicalHardwareModel], ABC): ...


class BasePhysicalModelLoader(BaseModelLoader[PydPhysicalHardwareModel], ABC): ...



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from enum import Enum

from qat.utils.pydantic import NoExtraFieldsFrozenModel


class Scale(Enum):
    """
    SI unit scales.
    """

    NANO = "n"
    MICRO = "u"
    MILLI = "m"
    DEFAULT = ""
    KILO = "k"
    MEGA = "M"
    GIGA = "G"
    TERA = "T"


class Quantity(NoExtraFieldsFrozenModel):
    amount: float
    scale: Scale

    def __str__(self):
        return f"{self.amount} {self.scale.value}"


class Frequency(Quantity):
    def __str__(self):
        return super().__str__() + "Hz"


class Time(Quantity):
    def __str__(self):
        return super().__str__() + "s"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from typing import Any

from pydantic import Field

from qat.model.device import Qubit
from qat.utils.pydantic import NoExtraFieldsModel, ValidatedDict, ValidatedList


class QubitRegister(NoExtraFieldsModel):
    qubits: ValidatedList[Qubit] = Field(default_factory=lambda: ValidatedList[Qubit]())

    @property
    def contents(self):
        return list(self.qubits)

    def __repr__(self):
        return "QubitRegister: " + str(self.qubits)


class CregIndexValue(NoExtraFieldsModel):
    """
    Used to reference when we are looking at a particular index in a creg variable.
    """

    register_name: str
    index: int
    value: Any

    @property
    def variable(self):
        return f"{self.register_name}[{self.index}]"

    def __repr__(self):
        return self.variable


class BitRegister(NoExtraFieldsModel):
    bits: ValidatedList[CregIndexValue] = Field(
        default_factory=lambda: ValidatedList[CregIndexValue]()
    )

    @property
    def contents(self):
        return list(self.bits)

    def __repr__(self):
        return "BitRegister: " + str(self.bits)


class Registers(NoExtraFieldsModel):
    quantum: ValidatedDict[str, QubitRegister] = Field(
        default_factory=lambda: ValidatedDict[str, QubitRegister]()
    )
    classic: ValidatedDict[str, BitRegister] = Field(
        default_factory=lambda: ValidatedDict[str, BitRegister]()
    )

    def __repr__(self):
        return "quantum: " + str(self.quantum) + "\nclassic: " + str(self.classic)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from abc import ABC, abstractmethod

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class PassConcept(ABC):
    """Base class describing the abstraction of a pass.

    See :class:`PassManager`.
    """

    @abstractmethod
    def run(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs):
        pass


class PassModel(PassConcept):
    """Implements the polymorphic pass API. A wrapper for any object providing a :meth:`run`
    method that accepts an argument as well as a :class:`ResultManager` and
    :class:`MetricsManager`.

    See :class:`PassManager`.
    """

    def __init__(self, pass_obj):
        self._pass = pass_obj

    def run(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs):
        """
        :param ir: Argument to pass to the object.
        :param res_mgr:
        :param met_mgr:
        """
        return self._pass.run(ir, res_mgr, met_mgr, *args, **kwargs)


class PassInfoMixin(ABC):
    """Base mixin specifying pass identification mechanism. A pass has an :attr:`id` and
    :attr:`name`.

    Such identification is loosely specified as it related closely to how passes and their
    results are managed throughout their lifecycle.

    See :class:`PassManager`.
    """

    def id(self):
        pass

    def name(self):
        pass


class AnalysisPass(PassInfoMixin):
    """Base class of all passes that compute some form of analysis on the input argument,
    with the input IR left intact."""

    def run(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs):
        pass


class TransformPass(PassInfoMixin):
    """Base class for all passes that mutates the IR.

    It is expected that child classes do not alter the type of the IR, but only alter its
    contents.
    """

    def run(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs):
        pass


class ValidationPass(PassInfoMixin):
    """Base class for all passes that verify or run some form of check (i.e semantics,
    legalisation, ...) on the IR.

    Behaviour is loose as to what the result of the validation would be: either a result or
    an error raised. It can change according to circumstances.
    """

    def run(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs):
        pass


class LoweringPass(PassInfoMixin):
    """Base case for all passes that modify the IR, instrinsically changing its type and
    structure.

    Acts as insulation between passes that that expect to see the IR in some
    given format.
    """

    def run(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs):
        pass


class PassManager(PassInfoMixin):
    """Contains a sequence of passes.

    Represents a sequential composite of passes, which can be composites themselves.
    Although not explicitly specified, the pass manager is best modelled as a DAG that acts
    as a pass itself. In doing so, it runs a sequence of (composite or leaf) passes over the
    input and aggregates results from them.

    Result aggregation (in general) can be achieved via return values or side effects. We
    adopted the latter approach where a :class:`ResultManager` is passed in as argument. The
    rationale behind this choice is two-fold:

    #. Allow passes to peek, use, or potentially invalidate results via the ResultManager
       without burdening their parent :class:`PassManager`.
    #. Let the passes and results define their dependency graphs on the fly which saves us
       from explicitly building and maintaining them. Here it's also a possible direction to
       let passes lazily declare, create, and run their dependencies as they get discovered.

    Result aggregation would also potentially involve cache handling as described in the
    :class:`ResultManager`. This calls for proper pass identification, cache invalidation,
    and pass/result cycle detection techniques.

    Example of a DAG:

    (PM1 A B (PM2 C D ) E F) can be thought of as:

    .. code-block::

                <---PM2--->

                   --> C
        A --> B --|    ↓
                   --> D --> E --> F

        <-----------PM1------------>

    Although the notation of a given (sub-) DAG looks sequential, passes within can run
    sequentially or in parallel depending on their inter-dependencies.

    Notes thus far describe an ideal fully-fledged PassManager. However, today's needs are
    very simple as we barely have visible features of modular quantum compilation workflow.
    It is therefore wise to keep the implementation light and basic where component passes
    run and register their own results within the :class:`ResultManager` passed in as
    argument. Another reason is the possibility in the near future to adopt and tap into
    existing and mature pass pipeline infrastructure which saves us from reinventing or
    rediscovering the same concepts already covered by the opensource community.
    """

    def __init__(self):
        self.passes: list[PassModel] = []

    def run(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs):
        """
        :param arg: Argument to pass to the object.
        :param res_mgr:
        :param met_mgr:
        """
        for p in self.passes:
            ir = p.run(ir, res_mgr, met_mgr, *args, **kwargs)
        return ir

    def add(self, pass_obj):
        """
        Add a pass to the pass manager.

        This can be achieved by either using :code:`pass_mgr.add(pass)`, or
        :code:`pass_mgr | pass`.

        :param pass_obj:
        """
        self.passes.append(PassModel(pass_obj))
        return self

    def __or__(self, pass_obj):
        return self.add(pass_obj)


class InvokerMixin(ABC):
    """Useful for compilation (global design-wise) stages/phases that are meant to invoke
    some arbitrary formation of passes in the form of a pipeline.

    Organisation, registering, visibility, and discovery of pipelines w.r.t the global
    quantum compilation workflow is hard to pin down early on in this design partly because
    today's needs are trivial and the traditional "hand-me-down" cascade-like compilation
    stages is likely to shift as QAT matures.

    However, we'll start by specifying that an invoker builds and runs its own pipeline,
    validates its analyses results, and uses those results for its purpose.
    """

    @abstractmethod
    def build_pass_pipeline(self, *args, **kwargs) -> PassManager:
        pass

    def run_pass_pipeline(
        self, arg, res_mgr: ResultManager, met_mgr: MetricsManager, *args, **kwargs
    ):
        return self.build_pass_pipeline(*args, **kwargs).run(
            arg, res_mgr, met_mgr, *args, **kwargs
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import uuid as py_uuid
from random import Random

uuid_randomiser = Random()


def uuid():
    return py_uuid.UUID(int=uuid_randomiser.getrandbits(128), version=4)


uuid4 = uuid



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from pathlib import Path

from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.loaders.base import BasePhysicalModelLoader


class FileModelLoader(BasePhysicalModelLoader):
    def __init__(self, path: Path | str):
        self.path: Path = Path(path).expanduser()

    def load(self) -> PhysicalHardwareModel:
        blob = self.path.read_text()
        return PhysicalHardwareModel.model_validate_json(blob)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from qat.backend.qblox.codegen import QbloxBackend1, QbloxBackend2
from qat.backend.qblox.config.constants import QbloxTargetData
from qat.engines.qblox.execution import QbloxEngine
from qat.frontend import AutoFrontend
from qat.middleend import CustomMiddleend
from qat.pipelines.pipeline import Pipeline
from qat.pipelines.purr.qblox.compile import (
    backend_pipeline1,
    backend_pipeline2,
    middleend_pipeline1,
    middleend_pipeline2,
)
from qat.pipelines.purr.qblox.execute import get_results_pipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.backends.qblox.live import QbloxLiveHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime import SimpleRuntime
from qat.runtime.aggregator import QBloxAggregator

log = get_default_logger()


class QbloxPipeline1(UpdateablePipeline):
    """
    A pipeline that compiles programs using the :class:`QbloxBackend1` and executes
    them using the :class:`QbloxEngine`.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QbloxLiveHardwareModel,
        target_data: QbloxTargetData | None,
        engine: QbloxEngine = None,
    ) -> Pipeline:
        target_data = target_data if target_data is not None else QbloxTargetData.default()
        results_pipeline = get_results_pipeline(model)
        return Pipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontend(model),
            middleend=CustomMiddleend(
                model=model,
                pipeline=middleend_pipeline1(model=model, target_data=target_data),
            ),
            backend=QbloxBackend1(
                model=model,
                pipeline=backend_pipeline1(),
            ),
            runtime=SimpleRuntime(
                engine=engine,
                results_pipeline=results_pipeline,
                aggregator=QBloxAggregator(),
            ),
            name=config.name,
        )


class QbloxPipeline2(UpdateablePipeline):
    """
    A pipeline that compiles programs using the :class:`QbloxBackend2` and executes
    them using the :class:`QbloxEngine`.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QbloxLiveHardwareModel,
        target_data: QbloxTargetData | None,
        engine: QbloxEngine = None,
    ) -> Pipeline:
        target_data = target_data if target_data is not None else QbloxTargetData.default()
        results_pipeline = get_results_pipeline(model)
        return Pipeline(
            model=model,
            target_data=target_data,
            frontend=AutoFrontend(model),
            middleend=CustomMiddleend(
                model=model,
                pipeline=middleend_pipeline2(model=model, target_data=target_data),
            ),
            backend=QbloxBackend2(
                model=model,
                pipeline=backend_pipeline2(),
            ),
            runtime=SimpleRuntime(
                engine=engine,
                results_pipeline=results_pipeline,
                aggregator=QBloxAggregator(),
            ),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC, abstractmethod
from typing import Generic

import numpy as np

from qat.executables import Program


class NativeEngine(ABC, Generic[Program]):
    """:class:`NativeEngine` acts as an interface between some target machine and an
    executable. They are used to connect to the target machine (if applicable), and execute
    and return the results."""

    @abstractmethod
    def execute(self, program: Program, **kwargs) -> dict[str, np.ndarray]:
        """Executes a compiled instruction program and returns results that are processed
        according to the acquires.

        The engine is expected to return the results as a dictionary with the output
        variables as keys.
        """
        ...


class ConnectionMixin(ABC):
    """Specifies a connection requirement for a :class:`NativeEngine`.

    Engines that execute on live hardware or a remote simulator might require a connection
    to the target to be established. This class is used to mix in connection capabilities.
    """

    @property
    @abstractmethod
    def is_connected(self) -> bool: ...

    @abstractmethod
    def connect(self): ...

    @abstractmethod
    def disconnect(self): ...

    def __del__(self):
        if self.is_connected:
            self.disconnect()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import numpy as np
from pydantic import BaseModel, Field

from qat.utils.pydantic import FloatNDArray, IntNDArray


class PathData(BaseModel):
    """
    This object wraps the actual data as a list of samples, the number of averages performed
    by the hardware (if any), and whether the hw observed any out-of-range samples.
    """

    avg_cnt: int = None
    oor: bool = Field(alias="out-of-range", default=False)
    data: FloatNDArray = Field(default_factory=lambda: FloatNDArray([]))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.avg_cnt != other.avg_cnt:
            return False
        if self.oor != other.oor:
            return False
        if self.data.size != other.data.size or np.any(self.data != other.data):
            return False
        return True


class IntegData(BaseModel):
    """
    Path 0 refers to I while Path 1 refers to Q
    """

    path0: FloatNDArray = Field(default_factory=lambda: FloatNDArray([]))
    path1: FloatNDArray = Field(default_factory=lambda: FloatNDArray([]))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.path0.size != other.path0.size or np.any(self.path0 != other.path0):
            return False
        if self.path1.size != other.path1.size or np.any(self.path1 != other.path1):
            return False
        return True


class ScopeAcqData(BaseModel):
    """
    Path 0 refers to I while Path 1 refers to Q. Their lengths are statically equal
    to :class:`Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS`
    """

    path0: PathData = PathData()
    path1: PathData = PathData()


class BinnedAcqData(BaseModel):
    """
    Binned data is data that's been acquired and processed via different routes such as squared acquisition,
    weighed integration. Processing here refers to steps like averaging, rotation, and thresholding
    which are executed by the hardware.
    """

    avg_cnt: IntNDArray = Field(default_factory=lambda: IntNDArray([]))
    integration: IntegData = IntegData()
    threshold: FloatNDArray = Field(default_factory=lambda: FloatNDArray([]))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.avg_cnt.size != other.avg_cnt.size or np.any(self.avg_cnt != other.avg_cnt):
            return False
        if self.integration != other.integration:
            return False
        if self.threshold.size != other.threshold.size or np.any(
            self.threshold != other.threshold
        ):
            return False
        return True


class BinnedAndScopeAcqData(BaseModel):
    """
    The actual acquisition data, it represents the value associated with the key "acquisition"
    in the acquisition blob returned by Qblox. This object contains scope data and binned data.
    """

    bins: BinnedAcqData = BinnedAcqData()
    scope: ScopeAcqData = ScopeAcqData()


class Acquisition(BaseModel):
    """
    Represents a single acquisition. In Qblox terminology, this object contains scope, integrated, and threshold
    data all at once. It's up to the SW layer to pick up what it needs and adapt it to its flow.

    An acquisition contains is described by a name, index, and blob data represented by :class:`AcqData`
    """

    name: str | None = None
    index: int = None
    acquisition: BinnedAndScopeAcqData = BinnedAndScopeAcqData()

    def __add__(self, other: "Acquisition") -> "Acquisition":
        """
        Acquisition addition follows concatenation semantics such as the case for strings.
        A few important details that might be adjusted in the future:
            + Resulting scope_data.path0.avg_cnt is taken as the minimum of the two
                Reason for the underestimation is to remain conservative and on the safe side
                (Can raise if strictness is required)
            + Resulting scope_data.path0.oor follows "AND" semantics
        """

        if not isinstance(other, Acquisition):
            raise TypeError(f"Can only add acquisitions, got {type(other)}")

        if self == Acquisition():
            return other.model_copy(deep=True)

        if other == Acquisition():
            return self.model_copy(deep=True)

        result = Acquisition()

        if self.index != other.index:
            raise ValueError(
                f"Expected the same index but got {self.index} != {other.index}"
            )
        result.index = self.index

        if self.name != other.name:
            raise ValueError(f"Expected the same name but got {self.name} != {other.name}")
        result.name = self.name

        scope_data1 = self.acquisition.scope
        scope_data2 = other.acquisition.scope
        scope_data = result.acquisition.scope

        scope_data.path0.avg_cnt = min(
            scope_data1.path0.avg_cnt or 0, scope_data2.path0.avg_cnt or 0
        )
        scope_data.path0.oor = scope_data1.path0.oor and scope_data2.path0.oor
        scope_data.path0.data = np.append(scope_data1.path0.data, scope_data2.path0.data)
        scope_data.path1.avg_cnt = min(
            scope_data1.path1.avg_cnt or 0, scope_data2.path1.avg_cnt or 0
        )
        scope_data.path1.oor = scope_data1.path1.oor and scope_data2.path1.oor
        scope_data.path1.data = np.append(scope_data1.path1.data, scope_data2.path1.data)

        bin_data1 = self.acquisition.bins
        bin_data2 = other.acquisition.bins
        bin_data = result.acquisition.bins

        bin_data.avg_cnt = np.append(bin_data1.avg_cnt, bin_data2.avg_cnt)
        bin_data.integration.path0 = np.append(
            bin_data1.integration.path0, bin_data2.integration.path0
        )
        bin_data.integration.path1 = np.append(
            bin_data1.integration.path1, bin_data2.integration.path1
        )
        bin_data.threshold = np.append(bin_data1.threshold, bin_data2.threshold)

        return result



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import numpy as np
from numpy.typing import NDArray
from pydantic import ConfigDict, field_validator

from qat.utils.pydantic import (
    CalibratableUnitInterval2x2Array,
    FrozenDict,
    NoExtraFieldsModel,
    QubitId,
)


class ReadoutMitigation(NoExtraFieldsModel):
    """
    :param linear: Maps each individual qubit to its <0/1> state given the <0/1> probabilistic
        error that happens on the device. Note that linear assumes no cross-talk effects and
        considers each qubit independent.
        linear = {
            <qubit_number>: NDArray(2, 2)
        }
        The matrix element (i, j) of the NDArray maps the probability p(i|j).
    :param m3_available: Flag to enable a runtime mitigation strategy that builds up the calibration
                        it needs at runtime. For more info https://github.com/Qiskit-Partners/mthree.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # TODO: COMPILER-706 linear readout mitigation currently a 2x2 matrix,
    #  we may want to change this to be a dictionary like in the old hardware model.
    linear: FrozenDict[QubitId, CalibratableUnitInterval2x2Array]
    matrix: NDArray | None = None
    m3_available: bool = False

    @field_validator("linear")
    @classmethod
    def validate_linear(cls, linear):
        for qubit_idx, qubit_map in linear.items():
            if not np.allclose(np.sum(qubit_map, axis=0), np.ones(qubit_map.shape[0])):
                raise ValueError(
                    f"Please provide a linear probability map for qubit {qubit_idx} where all probabilities p(i|{qubit_idx} sum to 1."
                )

        return linear

    @property
    def qubits(self):
        return list(self.linear.keys())


class ErrorMitigation(NoExtraFieldsModel):
    """
    A collection of error mitigation strategies. Currently, this holds a single mitigation strategy,
    but this can be expanded in the future as we add more error correction schemes.

    :param readout_mitigation: Linear readout mitigation.
    """

    readout_mitigation: ReadoutMitigation | None = None

    @property
    def is_enabled(self):
        return (
            True
            if self.readout_mitigation and len(self.readout_mitigation.linear)
            else False
        )

    @property
    def qubits(self):
        return self.readout_mitigation.qubits if self.readout_mitigation else []



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np
from pydantic import BaseModel, Field

from qat.executables import AbstractProgram
from qat.ir.measure import AcquireMode
from qat.utils.pydantic import ComplexNDArray


class PositionalAcquireData(BaseModel):
    """Contains the position to acquire from the readout signal, and the length of the
    readout."""

    output_variable: str
    position: int
    length: int
    mode: AcquireMode


class WaveformChannelData(BaseModel):
    """
    Contains the channel data for a :class:`WaveformV1Program`.

    Stores the waveforms and acqusitions needed for execution. No control flow is possible.

    :param buffer: The waveform to be sent at each sample.
    :type buffer: list[complex]
    :param baseband_frequency: The frequency to be set for the baseband.
    :type baseband_frequency: float | None
    :param acquires: Acquire information needed for readouts.
    :type acquires: list[AcquireData]
    """

    buffer: ComplexNDArray = Field(default_factory=lambda: ComplexNDArray([]))
    baseband_frequency: float | None = None
    acquires: list[PositionalAcquireData] = Field(default_factory=list)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        if self.acquires != other.acquires:
            return False
        if self.baseband_frequency != other.baseband_frequency:
            return False
        if not np.all(self.buffer == other.buffer):
            return False
        return True


class WaveformProgram(AbstractProgram):
    """Contains the information to execute a task using the waveform buffer channel hardware.

    Contains the buffers and acquire information for each channel, the repetition period for
    each shot and the number of shots to execute.

    :param channel_data: Contains the waveform buffers, acquisitions and baseband
        frequencies for each physical channel.
    :param repetition_time: The time each shot takes to execute.
    :param shots: The number of shots to be executed as part of the program.
    """

    channel_data: dict[str, WaveformChannelData]
    repetition_time: float
    shots: int

    @property
    def acquires(self) -> list[PositionalAcquireData]:
        acquires: list[PositionalAcquireData] = []
        for channel in self.channel_data.values():
            acquires.extend(channel.acquires)
        return acquires

    @property
    def acquire_shapes(self) -> dict[str, tuple[int, ...]]:
        acquire_shapes = {}
        for data in self.channel_data.values():
            for acquire in data.acquires:
                if acquire.mode == AcquireMode.SCOPE:
                    acquire_shapes[acquire.output_variable] = (acquire.length,)
                elif acquire.mode == AcquireMode.RAW:
                    acquire_shapes[acquire.output_variable] = (
                        self.shots,
                        acquire.length,
                    )
                else:
                    acquire_shapes[acquire.output_variable] = (self.shots,)
        return acquire_shapes



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
import json
from typing import TYPE_CHECKING

import compiler_config.serialiser as legacy_serialiser
import numpy as np

from qat.purr.compiler.devices import QuantumComponent
from qat.purr.compiler.instructions import Acquire

if TYPE_CHECKING:
    from qat.purr.compiler.hardware_models import QuantumHardwareModel


# Set of common methods so we don't have to add/remove the custom serializer.
def json_dumps(*args, serializable_types=None, **kwargs):
    kwargs.setdefault("cls", CustomQatJsonEncoder)
    return json.dumps(*args, serializable_types=serializable_types, **kwargs)


def json_loads(
    *args, serializable_types=None, model: "QuantumHardwareModel" = None, **kwargs
):
    kwargs.setdefault("cls", CustomQatJsonDecoder)
    return json.loads(*args, serializable_types=serializable_types, model=model, **kwargs)


def json_dump(*args, serializable_types=None, **kwargs):
    kwargs.setdefault("cls", CustomQatJsonEncoder)
    return json.dump(*args, serializable_types=serializable_types, **kwargs)


def json_load(
    *args, serializable_types=None, model: "QuantumHardwareModel" = None, **kwargs
):
    kwargs.setdefault("cls", CustomQatJsonDecoder)
    return json.load(*args, serializable_types=serializable_types, model=model, **kwargs)


class CustomQatJsonDecoder(legacy_serialiser.CustomJsonDecoder):
    def __init__(self, *args, model=None, **kwargs):
        self.model: "QuantumHardwareModel" = model
        super().__init__(*args, **kwargs)
        self.object_hook = self.default

    def default(self, obj):
        if not isinstance(obj, dict):
            return obj
        elif obj.get("type", False) == "numpyarray":
            if "value" in obj:
                array = np.frombuffer(
                    bytearray.fromhex(obj["value"]), dtype=np.dtype(obj["dtype"])
                )
                return array.reshape(obj["shape"])
            return np.array(obj["list"])

        # Components are objects directly related to hardware and you never want to serialize them, so we re-link
        # upon deserialization.
        component_id = obj.get("$component_id")
        if component_id is not None:
            if self.model is None:
                raise ValueError(
                    "Attempted to deserialize object that requires re-linking to a "
                    "hardware model and we have no hardware model."
                )

            return self.model.get_device(component_id)

        return super().default(obj)


class CustomQatJsonEncoder(legacy_serialiser.CustomJSONEncoder):
    """
    It is a customised JSON encoder, which allows the serialization of the more complex
    objects.

    There are four major cases, based on the provided object to be serialized:

    - if the type of the object is supported by the default :class:`JSONEncoder`, than
      the default method is used.
    - if the class of the object is a :class:`dataclass`, then the serialization will
      contain the name of the type, ``dataclass`` flag in order to help at the
      deserialization, and the dictionary of the fields.
    - if the object is none from the above, then the type name is saved, and the
      interior data using ``__dict__``.
    - if an exception is encountered from any cases from above (e.g. ``__dict__`` is not
      available in case of complex numbers), then the type name is saved, and the data
      is the string representation of the object.
    """

    def default(self, obj):
        # TODO: Acquire is a special wrapper component, not an actual component. Have a few too many special-cases
        #   for it now, think about reverting its special status.
        if isinstance(obj, QuantumComponent) and not isinstance(obj, Acquire):
            return {"$component_id": obj.full_id()}
        elif isinstance(obj, np.ndarray):
            return {
                "type": "numpyarray",
                "dtype": obj.dtype.name,
                "shape": obj.shape,
                "value": obj.tobytes().hex(),
            }
        return super().default(obj)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from qat.backend.qblox.config.constants import QbloxTargetData
from qat.core.pass_base import PassManager
from qat.engines.qblox.execution import QbloxEngine
from qat.pipelines.pipeline import ExecutePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.backends.qblox.live import QbloxLiveHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime import SimpleRuntime
from qat.runtime.aggregator import QBloxAggregator
from qat.runtime.passes.purr.analysis import IndexMappingAnalysis
from qat.runtime.passes.transform import (
    AssignResultsTransform,
    ErrorMitigation,
    InlineResultsProcessingTransform,
    QBloxAcquisitionPostProcessing,
    ResultTransform,
)

log = get_default_logger()


def get_results_pipeline(model: QuantumHardwareModel) -> PassManager:
    """Factory for creating the default results pipeline.

    :param model: The quantum hardware model to use for the pipeline.
    """

    return (
        PassManager()
        | QBloxAcquisitionPostProcessing()
        | InlineResultsProcessingTransform()
        | AssignResultsTransform()
        | ResultTransform()
        | IndexMappingAnalysis(model)
        | ErrorMitigation(model)
    )


class QbloxExecutePipeline(UpdateablePipeline):
    """A pipeline that executes :class:`Executable <qat.executable.Executable>`s with
    :class:`QbloxProgram <qat.backend.qblox.execution.QbloxProgram>`s
    packages using the :class:`QbloxLiveEngineAdapter`.

    .. warning::

        This pipeline is for execution purposes only and does not compile programs. Please
        select an appropriate compilation pipeline if you wish to compile programs before
        execution.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QbloxLiveHardwareModel,
        target_data: QbloxTargetData | None,
        engine: QbloxEngine = None,
    ) -> ExecutePipeline:
        target_data = target_data if target_data is not None else QbloxTargetData.default()
        results_pipeline = get_results_pipeline(model)
        return ExecutePipeline(
            model=model,
            target_data=target_data,
            runtime=SimpleRuntime(
                engine=engine,
                results_pipeline=results_pipeline,
                aggregator=QBloxAggregator(),
            ),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from collections import defaultdict

import networkx as nx


def get_connected_subgraphs(edges: list[tuple[int, int]]):
    """
    Given a list of edges in a graph, which can be disconnected,
    construct the connected subgraph(s) within a given graph.
    """
    G = nx.DiGraph(edges)

    subgraphs_nodes = []
    subgraphs_edges = []
    for subgr_nodes in nx.weakly_connected_components(G):
        subgraphs_nodes.append(subgr_nodes)
        subgraphs_edges.append(list(G.subgraph(subgr_nodes).edges()))

    return subgraphs_nodes, subgraphs_edges


def convert_edges_into_connectivity_dict(edges: list[tuple[int, int]]):
    connectivity = defaultdict(set)
    for edge in edges:
        connectivity[edge[0]].add(edge[1])
        connectivity[edge[1]].add(edge[0])

    return connectivity


def generate_cyclic_connectivity(n_nodes: int):
    G = nx.cycle_graph(n_nodes)
    return convert_edges_into_connectivity_dict(list(G.edges()))


def generate_complete_connectivity(n_nodes: int):
    G = nx.complete_graph(n_nodes)
    return convert_edges_into_connectivity_dict(list(G.edges()))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qiskit_aer.noise import NoiseModel

from qat.model.loaders.purr.base import BaseLegacyModelLoader
from qat.purr.backends.echo import Connectivity
from qat.purr.backends.qiskit_simulator import (
    QiskitHardwareModel as LegacyQiskitHardwareModel,
    get_default_qiskit_hardware,
)


class QiskitModelLoader(BaseLegacyModelLoader):
    def __init__(
        self,
        qubit_count: int = 20,
        noise_model: NoiseModel | None = None,
        strict_placement: bool = True,
        connectivity: Connectivity | list[(int, int)] | None = None,
    ):
        self.qubit_count = qubit_count
        self.noise_model = noise_model
        self.strict_placement = strict_placement
        self.connectivity = connectivity

    def load(self) -> LegacyQiskitHardwareModel:
        return get_default_qiskit_hardware(
            qubit_count=self.qubit_count,
            noise_model=self.noise_model,
            strict_placement=self.strict_placement,
            connectivity=self.connectivity,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from abc import abstractmethod
from collections import defaultdict
from functools import reduce
from itertools import groupby
from typing import Dict, List

from compiler_config.config import InlineResultsProcessing
from more_itertools import partition

from qat.purr.backends.live import LiveDeviceEngine, LiveHardwareModel
from qat.purr.backends.live_devices import ControlHardware
from qat.purr.backends.qblox.acquisition import Acquisition
from qat.purr.backends.qblox.analysis_passes import (
    BindingPass,
    QbloxLegalisationPass,
    TILegalisationPass,
    TriagePass,
    TriageResult,
)
from qat.purr.backends.qblox.codegen import NewQbloxEmitter, QbloxEmitter, QbloxPackage
from qat.purr.backends.qblox.transform_passes import (
    DesugaringPass,
    RepeatSanitisation,
    ReturnSanitisation,
    ScopeSanitisation,
)
from qat.purr.backends.utilities import software_post_process_linear_map_complex_to_real
from qat.purr.compiler.execution import (
    DeviceInjectors,
    _binary_average,
    _numpy_array_to_list,
)
from qat.purr.compiler.instructions import (
    AcquireMode,
    DeviceUpdate,
    IndexAccessor,
    Instruction,
    PostProcessType,
    ProcessAxis,
    Variable,
)
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt
from qat.purr.compiler.runtime import NewQuantumRuntime
from qat.purr.core.metrics_base import MetricsManager
from qat.purr.core.pass_base import InvokerMixin, PassManager
from qat.purr.core.result_base import ResultManager
from qat.purr.utils.logger import get_default_logger
from qat.purr.utils.logging_utils import log_duration

log = get_default_logger()


class QbloxLiveHardwareModel(LiveHardwareModel):
    def __init__(self, control_hardware: ControlHardware = None):
        super().__init__(control_hardware)
        self._reverse_coercion()

    def create_engine(self, startup_engine: bool = True):
        return QbloxLiveEngineAdapter(self, startup_engine)

    def create_runtime(self, existing_engine=None):
        existing_engine = existing_engine or self.create_engine()
        return NewQuantumRuntime(existing_engine)

    def _reverse_coercion(self):
        """
        HwM pickler coerces primitive types into strings. This is a temporary
        workaround to support in-memory, pickled, and un-pickled dict key types
        of the qblox config
        """
        for channel in self.physical_channels.values():
            config = channel.config
            config.sequencers = {int(k): v for k, v in config.sequencers.items()}


class AbstractQbloxLiveEngine(LiveDeviceEngine, InvokerMixin):
    def startup(self):
        if self.model.control_hardware is None:
            raise ValueError("Please add a control hardware first!")
        self.model.control_hardware.connect()

    def shutdown(self):
        if self.model.control_hardware is None:
            raise ValueError("Please add a control hardware first!")
        self.model.control_hardware.disconnect()

    def optimize(self, instructions):
        pass

    def validate(self, instructions: List[Instruction]):
        pass

    def _process_results(self, results, triage_result: TriageResult):
        """
        Process any software-driven results transformation, such as taking a raw
        waveform result and turning it into a bit, or something else.
        """

        rp_map = triage_result.rp_map

        for inst in rp_map.values():
            target_values = results.get(inst.variable, None)
            if target_values is None:
                raise ValueError(f"Variable {inst.variable} not found in results output.")

            if (
                InlineResultsProcessing.Raw in inst.results_processing
                and InlineResultsProcessing.Binary in inst.results_processing
            ):
                raise ValueError(
                    f"Raw and Binary processing attempted to be applied "
                    f"to {inst.variable}. Only one should be selected."
                )

            # Strip numpy arrays if we're set to do so.
            if InlineResultsProcessing.NumpyArrays not in inst.results_processing:
                target_values = _numpy_array_to_list(target_values)

            # Transform to various formats if required.
            if InlineResultsProcessing.Binary in inst.results_processing:
                target_values = _binary_average(target_values)

            results[inst.variable] = target_values

        return results

    def _process_assigns(self, results, triage_result: TriageResult):
        """
        As assigns are classical instructions they are not processed as a part of the
        quantum execution (right now).
        Read through the results dictionary and perform the assigns directly, return the
        results.
        """

        def recurse_arrays(results_map, value):
            """Recurse through assignment lists and fetch values in sequence."""
            if isinstance(value, List):
                return [recurse_arrays(results_map, val) for val in value]
            elif isinstance(value, Variable):
                if value.name not in results_map:
                    raise ValueError(
                        f"Attempt to assign variable that doesn't exist {value.name}."
                    )

                if isinstance(value, IndexAccessor):
                    return results_map[value.name][value.index]
                else:
                    return results_map[value.name]
            else:
                return value

        assigns = triage_result.assigns
        ret_inst = next(iter(triage_result.returns))
        assigned_results = dict(results)
        for assign in assigns:
            assigned_results[assign.name] = recurse_arrays(assigned_results, assign.value)

        return {key: assigned_results[key] for key in ret_inst.variables}

    @staticmethod
    def combine_playbacks(playbacks: Dict[str, List[Acquisition]]):
        """
        Combines acquisition objects from multiple acquire instructions in multiple readout targets.
        Notice that :meth:`groupby` preserves (original) relative order, which makes it honour
        the (sequential) lexicographical order of the loop nest:

        playback[target]["acq_0"] contains (potentially) a list of acquisitions collected in the same
        order as the order in which the packages were sent to the FPGA.

        Although acquisition names are enough for unicity in practice, the playback's structure
        distinguishes different (multiple) acquisitions per readout target, thus making it more robust.
        """

        playback: Dict[str, Dict[str, Acquisition]] = {}
        for pulse_channel_id, acquisitions in playbacks.items():
            acquisitions.sort(key=lambda acquisition: acquisition.name)
            groups_by_name = groupby(acquisitions, lambda acquisition: acquisition.name)
            playback[pulse_channel_id] = {
                name: reduce(
                    lambda acq1, acq2: acq1 + acq2,
                    acqs,
                    Acquisition(),
                )
                for name, acqs in groups_by_name
            }

        return playback

    def process_playback(
        self, playback: Dict[str, Dict[str, Acquisition]], res_mgr: ResultManager
    ):
        """
        Now that the combined playback is ready, we can compute and process results as required
        by customers. This requires loop nest information as well as post-processing and shaping
        requirements.
        """

        triage_result = res_mgr.lookup_by_type(TriageResult)
        acquire_map = triage_result.acquire_map
        sweeps = triage_result.sweeps
        repeats = triage_result.repeats
        pp_map = triage_result.pp_map
        sweep_counts = list(len(next(iter(s.variables.values()))) for s in sweeps)
        repeat_counts = list(r.repeat_count for r in repeats)

        results = {}
        for pulse_channel_id, acquisitions in playback.items():
            target = self.model.get_pulse_channel_from_id(pulse_channel_id)
            acquires = acquire_map[target]
            for name, acquisition in acquisitions.items():
                scope_data = acquisition.acquisition.scope
                integ_data = acquisition.acquisition.bins.integration
                thrld_data = acquisition.acquisition.bins.threshold

                acquire = next((inst for inst in acquires if inst.output_variable == name))
                if acquire.mode in [AcquireMode.SCOPE, AcquireMode.RAW]:
                    repeat_counts.clear()
                    response = scope_data.path0.data + 1j * scope_data.path1.data
                elif acquire.mode == AcquireMode.INTEGRATOR:
                    response = integ_data.path0 + 1j * integ_data.path1
                else:
                    raise ValueError(f"Unrecognised acquire mode {acquire.mode}")

                post_procs, axes = pp_map[name], {}
                for pp in post_procs:
                    if (
                        pp.process == PostProcessType.MEAN
                        and ProcessAxis.SEQUENCE in pp.axes
                    ):
                        repeat_counts.clear()
                    elif pp.process == PostProcessType.LINEAR_MAP_COMPLEX_TO_REAL:
                        response, _ = software_post_process_linear_map_complex_to_real(
                            pp.args, response, axes
                        )
                    elif pp.process == PostProcessType.DISCRIMINATE:
                        # f : {0, 1} ----> {-1, 1}
                        #       x    |---> 1 - 2x
                        response = 1 - 2 * thrld_data

                loop_nest_shape = tuple((sweep_counts or [1]) + (repeat_counts or [-1]))
                response = response.reshape(loop_nest_shape)
                if sweep_counts:
                    response = response.squeeze()
                results[name] = response

        results = self._process_results(results, triage_result)
        results = self._process_assigns(results, triage_result)

        return results

    @abstractmethod
    def invoke_backend(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager): ...

    def execute_packages(self, packages) -> Dict[str, Dict[str, Acquisition]]: ...

    def _common_execute(self, builder, interrupt: Interrupt = NullInterrupt()):
        self._model_exists()
        res_mgr = ResultManager()
        met_mgr = MetricsManager()

        with log_duration("Codegen run in {} seconds."):
            packages = self.invoke_backend(builder, res_mgr, met_mgr)

        with log_duration("QPU returned results in {} seconds."):
            playback = self.execute_packages(packages)
            results = self.process_playback(playback, res_mgr)
            return results


class QbloxLiveEngine1(AbstractQbloxLiveEngine):
    """
    Legacy engine that runs a loop nest statically. It pregenerates packages for each iteration and runs
    them in the same lexicographical order. This generates lots of playbacks that are then combined later.

    See QbloxBackend1
    """

    def build_pass_pipeline(self, *args, **kwargs):
        return (
            PassManager()
            | RepeatSanitisation(self.model)
            | ScopeSanitisation()
            | ReturnSanitisation()
            | DesugaringPass()
            | TriagePass()
            | BindingPass()
        )

    def invoke_backend(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager):
        self.run_pass_pipeline(ir, res_mgr, met_mgr)
        return QbloxEmitter().emit_packages(ir, res_mgr, met_mgr)

    def execute_packages(self, iter2packages):
        playbacks: Dict[str, List[Acquisition]] = defaultdict(list)
        for packages in iter2packages.values():
            self.model.control_hardware.set_data(packages)
            playback = self.model.control_hardware.start_playback(None, None)
            for pulse_channel_id, acquisitions in playback.items():
                playbacks[pulse_channel_id] += acquisitions

        return self.combine_playbacks(playbacks)


class QbloxLiveEngine2(AbstractQbloxLiveEngine):
    """
    Not all algorithms fall within Q1's capabilities. While this remains an analysis issue we saw great
    flexibility in forking out this engine to allow R&D without compromising existing execution environment.

    Unlike vanilla QbloxLiveEngine, this engine does not use static iteration and injection mechanism.
    It leverages Q1's programming model to accelerate a handful of pulse-level algorithms.

    See QbloxBackend2
    """

    def build_pass_pipeline(self, *args, **kwargs):
        return (
            PassManager()
            | RepeatSanitisation(self.model)
            | ScopeSanitisation()
            | ReturnSanitisation()
            | DesugaringPass()
            | TriagePass()
            | BindingPass()
            | TILegalisationPass()
            | QbloxLegalisationPass()
        )

    def invoke_backend(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager):
        # TODO - A skeptical usage of DeviceInjectors on static device updates
        # TODO - Figure out what they mean w/r to scopes and control flow
        remaining, static_dus = partition(
            lambda inst: isinstance(inst, DeviceUpdate)
            and not isinstance(inst.value, Variable),
            ir.instructions,
        )
        remaining, static_dus = list(remaining), list(static_dus)

        ir.instructions = remaining
        injectors = DeviceInjectors(static_dus)

        try:
            injectors.inject()
            res_mgr = res_mgr or ResultManager()
            met_mgr = met_mgr or MetricsManager()
            self.run_pass_pipeline(ir, res_mgr, met_mgr)
            return NewQbloxEmitter().emit_packages(ir, res_mgr, met_mgr)
        finally:
            injectors.revert()

    def execute_packages(self, packages: List[QbloxPackage]):
        self.model.control_hardware.set_data(packages)
        playbacks: Dict[str, List[Acquisition]] = (
            self.model.control_hardware.start_playback(None, None)
        )

        return self.combine_playbacks(playbacks)


class QbloxLiveEngineAdapter(AbstractQbloxLiveEngine):
    """
    A manual adapter of the new and legacy engines. Users can switch on and off HW acceleration by using
    the `enable_hax` flag.

    There should be a proper hybrid, dynamic, JIT-like engine in the future that can leverage the HW to accelerate
    programs whenever possible.
    """

    def invoke_backend(self, ir, res_mgr: ResultManager, met_mgr: MetricsManager):
        pass

    def build_pass_pipeline(self, *args, **kwargs) -> PassManager:
        pass

    model: QbloxLiveHardwareModel

    def __init__(
        self,
        model: QbloxLiveHardwareModel,
        startup_engine: bool = True,
        enable_hax=False,
    ):
        super().__init__(model, startup_engine)
        self._legacy_engine = QbloxLiveEngine1(model, False)
        self._new_engine = QbloxLiveEngine2(model, False)
        self.enable_hax = enable_hax

    def _common_execute(self, builder, interrupt: Interrupt = NullInterrupt()):
        if self.enable_hax:
            return self._new_engine._common_execute(builder, interrupt)
        return self._legacy_engine._common_execute(builder, interrupt)
